from django.utils.html import escape

from wagtail.core.models import Page
from wagtail.core.rich_text import LinkHandler


class PageLinkHandler(LinkHandler):
    identifier = 'page'

    @staticmethod
    def get_model():
        return Page

    @classmethod
    def get_instance(cls, attrs):
        return super().get_instance(attrs).specific

    @classmethod
    def expand_db_attributes(cls, attrs):
        try:
            page = cls.get_instance(attrs)
            return '<a href="%s">' % escape(page.specific.url)
        except Page.DoesNotExist:
            return "<a>"



from django import forms
from django.utils.translation import ugettext as _
from django.utils.translation import ugettext_lazy


class SearchForm(forms.Form):
    def __init__(self, *args, **kwargs):
        placeholder = kwargs.pop('placeholder', _("Search"))
        super().__init__(*args, **kwargs)
        self.fields['q'].widget.attrs = {'placeholder': placeholder}

    q = forms.CharField(label=ugettext_lazy("Search term"), widget=forms.TextInput())



# -*- coding: utf-8 -*-

# AUTOMATICALLY GENERATED FILE, DO NOT EDIT

tz_locations = {
    'Brazil': u'Brazil',
    'Canada': u'Canada',
    'Asia': u'Asia',
    'America/Indiana/Tell_City': u'Tell City',
    'America/North_Dakota/New_Salem': u'New Salem',
    'US': u'United States',
    'Indiana': u'Indiana',
    'Argentina': u'Argentina',
    'Europe': u'Europe',
    'Australia': u'Australia',
    'America/Indiana/Vevay': u'Vevay',
    'Arctic': u'Arctic',
    'America/Indiana/Marengo': u'Marengo',
    'Atlantic': u'Atlantic',
    'Chile': u'Chile',
    'America/Indiana/Vincennes': u'Vincennes',
    'America': u'America',
    'America/North_Dakota/Center': u'Center',
    'Kentucky': u'Kentucky',
    'Etc': u'Etc',
    'North_Dakota': u'North Dakota',
    'America/Indiana/Knox': u'Knox',
    'America/Kentucky/Monticello': u'Monticello',
    'America/North_Dakota/Beulah': u'Beulah',
    'America/Indiana/Petersburg': u'Petersburg',
    'Mexico': u'Mexico',
    'Africa': u'Africa',
    'Pacific': u'Pacific',
    'Antarctica': u'Antarctica',
    'Indian': u'Indian',
    'America/Indiana/Winamac': u'Winamac',
}

tz_cities = {
    'Atlantic/Canary': u'Canary',
    'Australia/Melbourne': u'Melbourne',
    'Etc/GMT+9': u'GMT+9',
    'Etc/GMT+8': u'GMT+8',
    'Europe/Lisbon': u'Lisbon',
    'Etc/GMT+3': u'GMT+3',
    'Etc/GMT+2': u'GMT+2',
    'Etc/GMT+1': u'GMT+1',
    'Etc/GMT+0': u'GMT+0',
    'America/Nipigon': u'Nipigon',
    'Etc/GMT+6': u'GMT+6',
    'Etc/GMT+5': u'GMT+5',
    'America/Miquelon': u'Miquelon',
    'Pacific/Wallis': u'Wallis',
    'America/Fort_Nelson': u'Fort Nelson',
    'Antarctica/Davis': u'Davis',
    'America/Coral_Harbour': u'Atikokan',
    'Asia/Dhaka': u'Dhaka',
    'America/St_Lucia': u'St. Lucia',
    'Canada/Newfoundland': u'Newfoundland',
    'Asia/Kashgar': u'Kashgar',
    'America/Phoenix': u'Phoenix',
    'Europe/Kaliningrad': u'Kaliningrad',
    'Etc/GMT+7': u'GMT+7',
    'America/Mazatlan': u'Mazatlan',
    'Arctic/Longyearbyen': u'Longyearbyen',
    'Europe/Guernsey': u'Guernsey',
    'GB': u'GB',
    'Europe/Paris': u'Paris',
    'Europe/Stockholm': u'Stockholm',
    'Pacific/Fiji': u'Fiji',
    'Pacific/Apia': u'Apia',
    'Etc/GMT+4': u'GMT+4',
    'Pacific/Pago_Pago': u'Pago Pago',
    'Asia/Rangoon': u'Rangoon',
    'America/Mexico_City': u'Mexico City',
    'America/Ensenada': u'Ensenada',
    'America/Puerto_Rico': u'Puerto Rico',
    'Indian/Mauritius': u'Mauritius',
    'Europe/Berlin': u'Berlin',
    'Europe/Zurich': u'Zurich',
    'Africa/Casablanca': u'Casablanca',
    'Antarctica/Macquarie': u'Macquarie Island',
    'Asia/Krasnoyarsk': u'Krasnoyarsk',
    'Australia/Canberra': u'Canberra',
    'Atlantic/Bermuda': u'Bermuda',
    'Asia/Ujung_Pandang': u'Ujung Pandang',
    'America/Araguaina': u'Araguaina',
    'Asia/Tehran': u'Tehran',
    'Asia/Saigon': u'Ho Chi Minh',
    'Asia/Baku': u'Baku',
    'America/St_Barthelemy': u'St. Barthelemy',
    'America/Porto_Acre': u'Porto Acre',
    'America/Santarem': u'Santarem',
    'America/Argentina/Cordoba': u'Cordoba',
    'Libya': u'Libya',
    'America/Danmarkshavn': u'Danmarkshavn',
    'America/Scoresbysund': u'Ittoqqortoormiit',
    'America/Eirunepe': u'Eirunepe',
    'America/Caracas': u'Caracas',
    'Asia/Baghdad': u'Baghdad',
    'Africa/Monrovia': u'Monrovia',
    'Atlantic/Stanley': u'Stanley',
    'America/St_Vincent': u'St. Vincent',
    'Portugal': u'Portugal',
    'Europe/Tiraspol': u'Tiraspol',
    'America/Vancouver': u'Vancouver',
    'Asia/Ho_Chi_Minh': u'Ho Chi Minh',
    'Europe/Busingen': u'Busingen',
    'Etc/GMT0': u'GMT0',
    'Asia/Thimphu': u'Thimphu',
    'Africa/Ouagadougou': u'Ouagadougou',
    'America/Belize': u'Belize',
    'America/Port_of_Spain': u'Port of Spain',
    'Greenwich': u'Greenwich',
    'Asia/Tashkent': u'Tashkent',
    'Asia/Tokyo': u'Tokyo',
    'Pacific/Kiritimati': u'Kiritimati',
    'Australia/Sydney': u'Sydney',
    'Europe/Riga': u'Riga',
    'Asia/Dili': u'Dili',
    'Africa/Mbabane': u'Mbabane',
    'Asia/Oral': u'Oral',
    'Asia/Aden': u'Aden',
    'Europe/Isle_of_Man': u'Isle of Man',
    'Europe/Istanbul': u'Istanbul',
    'Asia/Magadan': u'Magadan',
    'Australia/Lindeman': u'Lindeman',
    'US/Michigan': u'Michigan',
    'Pacific/Galapagos': u'Galapagos',
    'America/Bogota': u'Bogota',
    'Africa/Asmara': u'Asmara',
    'America/Chicago': u'Chicago',
    'Pacific/Kwajalein': u'Kwajalein',
    'Australia/Broken_Hill': u'Broken Hill',
    'America/Cuiaba': u'Cuiaba',
    'Indian/Christmas': u'Christmas',
    'Asia/Jayapura': u'Jayapura',
    'Europe/Brussels': u'Brussels',
    'Asia/Chongqing': u'Chongqing',
    'Australia/NSW': u'NSW',
    'Canada/Pacific': u'Pacific',
    'America/Noronha': u'Noronha',
    'Europe/Podgorica': u'Podgorica',
    'Africa/Algiers': u'Algiers',
    'Africa/Harare': u'Harare',
    'Africa/Ndjamena': u'Ndjamena',
    'America/Costa_Rica': u'Costa Rica',
    'Europe/Ljubljana': u'Ljubljana',
    'Indian/Mayotte': u'Mayotte',
    'Asia/Phnom_Penh': u'Phnom Penh',
    'America/Managua': u'Managua',
    'America/Pangnirtung': u'Pangnirtung',
    'Etc/GMT+12': u'GMT+12',
    'America/Tijuana': u'Tijuana',
    'Pacific/Fakaofo': u'Fakaofo',
    'America/Adak': u'Adak',
    'America/Antigua': u'Antigua',
    'America/Indiana/Indianapolis': u'Indianapolis',
    'America/Argentina/La_Rioja': u'La Rioja',
    'Pacific/Tahiti': u'Tahiti',
    'Asia/Brunei': u'Brunei',
    'Europe/Zagreb': u'Zagreb',
    'America/Asuncion': u'Asuncion',
    'Europe/Vienna': u'Vienna',
    'Mexico/General': u'General',
    'Australia/Hobart': u'Hobart',
    'America/Juneau': u'Juneau',
    'America/Inuvik': u'Inuvik',
    'America/Ojinaga': u'Ojinaga',
    'Europe/Astrakhan': u'Astrakhan',
    'America/Montreal': u'Montreal',
    'Asia/Seoul': u'Seoul',
    'Brazil/East': u'East',
    'Indian/Comoro': u'Comoro',
    'Antarctica/Rothera': u'Rothera',
    'Hongkong': u'Hongkong',
    'Europe/Tallinn': u'Tallinn',
    'Indian/Mahe': u'Mahe',
    'Asia/Calcutta': u'Kolkata',
    'Australia/South': u'South',
    'America/Argentina/Jujuy': u'Jujuy',
    'Asia/Bishkek': u'Bishkek',
    'America/Creston': u'Creston',
    'America/Martinique': u'Martinique',
    'Asia/Singapore': u'Singapore',
    'PRC': u'PRC',
    'Africa/Nairobi': u'Nairobi',
    'Asia/Thimbu': u'Thimbu',
    'America/Maceio': u'Maceio',
    'Africa/Cairo': u'Cairo',
    'Europe/Moscow': u'Moscow',
    'Antarctica/Palmer': u'Palmer',
    'Asia/Ulaanbaatar': u'Ulaanbaatar',
    'America/Rainy_River': u'Rainy River',
    'Africa/Kampala': u'Kampala',
    'Asia/Colombo': u'Colombo',
    'Australia/Adelaide': u'Adelaide',
    'America/Cambridge_Bay': u'Cambridge Bay',
    'Africa/Luanda': u'Luanda',
    'Pacific/Chatham': u'Chatham',
    'America/Indiana/Winamac': u'Winamac, Indiana',
    'Asia/Dacca': u'Dacca',
    'America/Cordoba': u'Cordoba',
    'Asia/Tbilisi': u'Tbilisi',
    'Europe/Gibraltar': u'Gibraltar',
    'Asia/Karachi': u'Karachi',
    'Asia/Harbin': u'Harbin',
    'Australia/Lord_Howe': u'Lord Howe',
    'Etc/GMT-9': u'GMT-9',
    'Etc/GMT-8': u'GMT-8',
    'America/Bahia_Banderas': u'Bahia Banderas',
    'Etc/GMT-1': u'GMT-1',
    'Etc/GMT-0': u'GMT-0',
    'Etc/GMT-3': u'GMT-3',
    'Etc/GMT-2': u'GMT-2',
    'Etc/GMT-5': u'GMT-5',
    'Etc/GMT-4': u'GMT-4',
    'Etc/GMT-7': u'GMT-7',
    'Europe/Nicosia': u'Nicosia',
    'America/Boa_Vista': u'Boa Vista',
    'Eire': u'Eire',
    'America/Lima': u'Lima',
    'Indian/Reunion': u'Reunion',
    'Canada/Mountain': u'Mountain',
    'Asia/Srednekolymsk': u'Srednekolymsk',
    'CET': u'CET',
    'Europe/Belfast': u'Belfast',
    'America/Blanc-Sablon': u'Blanc-Sablon',
    'Australia/West': u'West',
    'Antarctica/Syowa': u'Syowa',
    'America/Jamaica': u'Jamaica',
    'Europe/Kiev': u'Kiev',
    'Europe/Budapest': u'Budapest',
    'Canada/Eastern': u'Eastern',
    'Pacific/Midway': u'Midway',
    'America/Goose_Bay': u'Goose Bay',
    'Australia/Victoria': u'Victoria',
    'Asia/Amman': u'Amman',
    'Asia/Sakhalin': u'Sakhalin',
    'Africa/Windhoek': u'Windhoek',
    'US/Pacific': u'Pacific',
    'America/Sitka': u'Sitka',
    'Asia/Katmandu': u'Kathmandu',
    'America/Guyana': u'Guyana',
    'Pacific/Pohnpei': u'Pohnpei',
    'America/Sao_Paulo': u'Sao Paulo',
    'Turkey': u'Turkey',
    'Australia/Yancowinna': u'Yancowinna',
    'America/Lower_Princes': u'Lower Prince’s Quarter',
    'Australia/Perth': u'Perth',
    'Africa/Djibouti': u'Djibouti',
    'Asia/Jakarta': u'Jakarta',
    'Asia/Pyongyang': u'Pyongyang',
    'EST5EDT': u'EST5EDT',
    'Africa/Johannesburg': u'Johannesburg',
    'Canada/East-Saskatchewan': u'East-Saskatchewan',
    'Asia/Tel_Aviv': u'Tel Aviv',
    'Antarctica/Troll': u'Troll',
    'Asia/Istanbul': u'Istanbul',
    'Asia/Irkutsk': u'Irkutsk',
    'Iran': u'Iran',
    'Australia/ACT': u'ACT',
    'Africa/Niamey': u'Niamey',
    'Australia/North': u'North',
    'America/Belem': u'Belem',
    'America/Indiana/Marengo': u'Marengo, Indiana',
    'Africa/Nouakchott': u'Nouakchott',
    'Asia/Ashkhabad': u'Ashkhabad',
    'Europe/Vilnius': u'Vilnius',
    'America/Cayenne': u'Cayenne',
    'Africa/Mogadishu': u'Mogadishu',
    'America/Kentucky/Monticello': u'Monticello, Kentucky',
    'America/Rio_Branco': u'Rio Branco',
    'America/Cancun': u'Cancun',
    'America/Havana': u'Havana',
    'Chile/EasterIsland': u'EasterIsland',
    'Pacific/Guam': u'Guam',
    'Pacific/Kosrae': u'Kosrae',
    'Universal': u'Universal',
    'US/Arizona': u'Arizona',
    'Atlantic/Azores': u'Azores',
    'Australia/Eucla': u'Eucla',
    'Asia/Shanghai': u'Shanghai',
    'US/Alaska': u'Alaska',
    'America/Godthab': u'Nuuk',
    'Asia/Beirut': u'Beirut',
    'Africa/Maputo': u'Maputo',
    'HST': u'HST',
    'Asia/Bahrain': u'Bahrain',
    'Asia/Ashgabat': u'Ashgabat',
    'Asia/Riyadh': u'Riyadh',
    'Atlantic/Faeroe': u'Faroe',
    'America/Montevideo': u'Montevideo',
    'America/Anguilla': u'Anguilla',
    'Asia/Damascus': u'Damascus',
    'America/North_Dakota/Center': u'Center, North Dakota',
    'UCT': u'UCT',
    'America/Indiana/Vevay': u'Vevay, Indiana',
    'Atlantic/St_Helena': u'St. Helena',
    'Iceland': u'Iceland',
    'Europe/Vatican': u'Vatican',
    'America/Indiana/Vincennes': u'Vincennes, Indiana',
    'Asia/Almaty': u'Almaty',
    'Australia/Queensland': u'Queensland',
    'America/Santo_Domingo': u'Santo Domingo',
    'Africa/Brazzaville': u'Brazzaville',
    'America/Nome': u'Nome',
    'Asia/Taipei': u'Taipei',
    'America/Yakutat': u'Yakutat',
    'W-SU': u'W-SU',
    'America/Argentina/Mendoza': u'Mendoza',
    'Zulu': u'Zulu',
    'Egypt': u'Egypt',
    'America/Tortola': u'Tortola',
    'Etc/UTC': u'UTC',
    'Brazil/Acre': u'Acre',
    'Africa/Asmera': u'Asmara',
    'Antarctica/Mawson': u'Mawson',
    'America/Buenos_Aires': u'Buenos Aires',
    'Asia/Kolkata': u'Kolkata',
    'Africa/Maseru': u'Maseru',
    'America/Atikokan': u'Atikokan',
    'America/Louisville': u'Louisville',
    'Pacific/Yap': u'Yap',
    'America/Santa_Isabel': u'Santa Isabel',
    'Asia/Kuching': u'Kuching',
    'Africa/Libreville': u'Libreville',
    'Africa/Freetown': u'Freetown',
    'Africa/Bissau': u'Bissau',
    'Europe/Samara': u'Samara',
    'Europe/Amsterdam': u'Amsterdam',
    'Europe/Tirane': u'Tirane',
    'Pacific/Saipan': u'Saipan',
    'Africa/Abidjan': u'Abidjan',
    'Europe/Zaporozhye': u'Zaporozhye',
    'America/Atka': u'Atka',
    'America/El_Salvador': u'El Salvador',
    'Europe/Madrid': u'Madrid',
    'Africa/Juba': u'Juba',
    'America/Santiago': u'Santiago',
    'America/Argentina/Buenos_Aires': u'Buenos Aires',
    'America/Argentina/San_Luis': u'San Luis',
    'Europe/Skopje': u'Skopje',
    'WET': u'WET',
    'America/Aruba': u'Aruba',
    'Navajo': u'Navajo',
    'America/Indianapolis': u'Indianapolis',
    'America/Regina': u'Regina',
    'Pacific/Truk': u'Chuuk',
    'Pacific/Chuuk': u'Chuuk',
    'Asia/Khandyga': u'Khandyga',
    'Pacific/Funafuti': u'Funafuti',
    'Canada/Atlantic': u'Atlantic',
    'America/Merida': u'Merida',
    'America/Guatemala': u'Guatemala',
    'Africa/Sao_Tome': u'Sao Tome',
    'Asia/Tomsk': u'Tomsk',
    'Asia/Makassar': u'Makassar',
    'Africa/Bujumbura': u'Bujumbura',
    'Europe/Chisinau': u'Chisinau',
    'Europe/Warsaw': u'Warsaw',
    'Asia/Yekaterinburg': u'Yekaterinburg',
    'US/Hawaii': u'Hawaii',
    'MET': u'MET',
    'Antarctica/Casey': u'Casey',
    'Pacific/Enderbury': u'Enderbury',
    'America/Thule': u'Thule',
    'America/St_Johns': u'St. John’s',
    'America/Moncton': u'Moncton',
    'Europe/Helsinki': u'Helsinki',
    'Atlantic/Cape_Verde': u'Cape Verde',
    'America/Tegucigalpa': u'Tegucigalpa',
    'Indian/Cocos': u'Cocos',
    'America/Boise': u'Boise',
    'America/Guadeloupe': u'Guadeloupe',
    'America/Nassau': u'Nassau',
    'Europe/Prague': u'Prague',
    'America/Halifax': u'Halifax',
    'Asia/Hovd': u'Hovd',
    'Canada/Saskatchewan': u'Saskatchewan',
    'America/Manaus': u'Manaus',
    'America/Rankin_Inlet': u'Rankin Inlet',
    'Etc/UCT': u'UCT',
    'GB-Eire': u'GB-Eire',
    'Etc/Greenwich': u'Greenwich',
    'Atlantic/Jan_Mayen': u'Jan Mayen',
    'America/North_Dakota/Beulah': u'Beulah, North Dakota',
    'America/Chihuahua': u'Chihuahua',
    'America/Iqaluit': u'Iqaluit',
    'America/Argentina/Rio_Gallegos': u'Rio Gallegos',
    'Pacific/Gambier': u'Gambier',
    'Europe/Volgograd': u'Volgograd',
    'Africa/Bamako': u'Bamako',
    'Asia/Novokuznetsk': u'Novokuznetsk',
    'Europe/Uzhgorod': u'Uzhgorod',
    'Africa/Banjul': u'Banjul',
    'Asia/Aqtau': u'Aqtau',
    'Pacific/Palau': u'Palau',
    'Africa/Malabo': u'Malabo',
    'Europe/Minsk': u'Minsk',
    'PST8PDT': u'PST8PDT',
    'America/Argentina/ComodRivadavia': u'ComodRivadavia',
    'Atlantic/Madeira': u'Madeira',
    'Pacific/Noumea': u'Noumea',
    'GMT0': u'GMT0',
    'Africa/Kinshasa': u'Kinshasa',
    'Europe/Malta': u'Malta',
    'US/Samoa': u'Samoa',
    'America/Argentina/Ushuaia': u'Ushuaia',
    'Asia/Chungking': u'Chungking',
    'Asia/Bangkok': u'Bangkok',
    'Pacific/Niue': u'Niue',
    'America/Catamarca': u'Catamarca',
    'America/Recife': u'Recife',
    'MST': u'MST',
    'Asia/Yerevan': u'Yerevan',
    'America/La_Paz': u'La Paz',
    'Poland': u'Poland',
    'Asia/Urumqi': u'Urumqi',
    'Africa/Lusaka': u'Lusaka',
    'US/Indiana-Starke': u'Indiana-Starke',
    'Pacific/Guadalcanal': u'Guadalcanal',
    'America/Yellowknife': u'Yellowknife',
    'Australia/LHI': u'LHI',
    'Asia/Vientiane': u'Vientiane',
    'Asia/Kuwait': u'Kuwait',
    'Africa/Conakry': u'Conakry',
    'Asia/Dubai': u'Dubai',
    'America/Argentina/Tucuman': u'Tucuman',
    'Asia/Chita': u'Chita',
    'Europe/Oslo': u'Oslo',
    'Australia/Currie': u'Currie',
    'America/St_Kitts': u'St. Kitts',
    'America/Panama': u'Panama',
    'America/Hermosillo': u'Hermosillo',
    'Asia/Hebron': u'Hebron',
    'America/Guayaquil': u'Guayaquil',
    'Asia/Kuala_Lumpur': u'Kuala Lumpur',
    'Europe/London': u'London',
    'America/Menominee': u'Menominee',
    'Asia/Kamchatka': u'Kamchatka',
    'Europe/Ulyanovsk': u'Ulyanovsk',
    'Asia/Vladivostok': u'Vladivostok',
    'America/Matamoros': u'Matamoros',
    'Brazil/DeNoronha': u'DeNoronha',
    'Asia/Qatar': u'Qatar',
    'Israel': u'Israel',
    'NZ-CHAT': u'NZ-CHAT',
    'Asia/Dushanbe': u'Dushanbe',
    'Asia/Yakutsk': u'Yakutsk',
    'Asia/Omsk': u'Omsk',
    'Africa/Bangui': u'Bangui',
    'UTC': u'UTC',
    'America/Paramaribo': u'Paramaribo',
    'Etc/GMT-11': u'GMT-11',
    'Etc/GMT-10': u'GMT-10',
    'Etc/GMT-13': u'GMT-13',
    'Etc/GMT-12': u'GMT-12',
    'Etc/GMT-14': u'GMT-14',
    'Pacific/Marquesas': u'Marquesas',
    'US/Central': u'Central',
    'Europe/Bratislava': u'Bratislava',
    'Asia/Anadyr': u'Anadyr',
    'America/New_York': u'New York',
    'Pacific/Norfolk': u'Norfolk',
    'CST6CDT': u'CST6CDT',
    'Pacific/Rarotonga': u'Rarotonga',
    'America/Dominica': u'Dominica',
    'Africa/Porto-Novo': u'Porto-Novo',
    'Asia/Samarkand': u'Samarkand',
    'America/Kentucky/Louisville': u'Louisville',
    'America/Toronto': u'Toronto',
    'America/Bahia': u'Bahia',
    'Indian/Maldives': u'Maldives',
    'Africa/Accra': u'Accra',
    'Antarctica/South_Pole': u'South Pole',
    'Asia/Muscat': u'Muscat',
    'America/Virgin': u'Virgin',
    'America/Edmonton': u'Edmonton',
    'Pacific/Wake': u'Wake',
    'America/Indiana/Tell_City': u'Tell City, Indiana',
    'Australia/Darwin': u'Darwin',
    'America/Whitehorse': u'Whitehorse',
    'America/Swift_Current': u'Swift Current',
    'Europe/Copenhagen': u'Copenhagen',
    'America/Fort_Wayne': u'Fort Wayne',
    'America/Montserrat': u'Montserrat',
    'America/Mendoza': u'Mendoza',
    'US/Mountain': u'Mountain',
    'Europe/Simferopol': u'Simferopol',
    'Africa/Blantyre': u'Blantyre',
    'America/Detroit': u'Detroit',
    'America/Shiprock': u'Shiprock',
    'America/Grenada': u'Grenada',
    'Atlantic/Faroe': u'Faroe',
    'America/Indiana/Petersburg': u'Petersburg, Indiana',
    'Asia/Kathmandu': u'Kathmandu',
    'Asia/Pontianak': u'Pontianak',
    'Jamaica': u'Jamaica',
    'Europe/Athens': u'Athens',
    'America/Port-au-Prince': u'Port-au-Prince',
    'America/Cayman': u'Cayman',
    'Etc/Universal': u'Universal',
    'Africa/Dar_es_Salaam': u'Dar es Salaam',
    'America/Curacao': u'Curacao',
    'Indian/Kerguelen': u'Kerguelen',
    'Africa/Khartoum': u'Khartoum',
    'Asia/Manila': u'Manila',
    'Africa/Lome': u'Lome',
    'America/Jujuy': u'Jujuy',
    'Africa/Douala': u'Douala',
    'EET': u'EET',
    'America/Argentina/San_Juan': u'San Juan',
    'America/North_Dakota/New_Salem': u'New Salem, North Dakota',
    'America/Kralendijk': u'Kralendijk',
    'Pacific/Port_Moresby': u'Port Moresby',
    'Europe/Jersey': u'Jersey',
    'Asia/Macao': u'Macao',
    'Europe/Andorra': u'Andorra',
    'ROK': u'ROK',
    'Cuba': u'Cuba',
    'Europe/Luxembourg': u'Luxembourg',
    'Pacific/Honolulu': u'Honolulu',
    'ROC': u'ROC',
    'America/St_Thomas': u'St. Thomas',
    'Pacific/Majuro': u'Majuro',
    'Asia/Hong_Kong': u'Hong Kong',
    'Asia/Macau': u'Macau',
    'MST7MDT': u'MST7MDT',
    'Europe/Belgrade': u'Belgrade',
    'Asia/Choibalsan': u'Choibalsan',
    'US/Eastern': u'Eastern',
    'Europe/Mariehamn': u'Mariehamn',
    'Antarctica/McMurdo': u'McMurdo',
    'America/Thunder_Bay': u'Thunder Bay',
    'America/Los_Angeles': u'Los Angeles',
    'Asia/Kabul': u'Kabul',
    'Indian/Antananarivo': u'Antananarivo',
    'Europe/Sarajevo': u'Sarajevo',
    'Atlantic/Reykjavik': u'Reykjavik',
    'Asia/Nicosia': u'Nicosia',
    'Etc/GMT+11': u'GMT+11',
    'Etc/GMT+10': u'GMT+10',
    'Pacific/Ponape': u'Pohnpei',
    'Japan': u'Japan',
    'Asia/Ulan_Bator': u'Ulan Bator',
    'Asia/Barnaul': u'Barnaul',
    'Kwajalein': u'Kwajalein',
    'Pacific/Tongatapu': u'Tongatapu',
    'America/Marigot': u'Marigot',
    'Pacific/Pitcairn': u'Pitcairn',
    'Pacific/Easter': u'Easter Island',
    'US/East-Indiana': u'East-Indiana',
    'Atlantic/South_Georgia': u'South Georgia',
    'Africa/El_Aaiun': u'El Aaiun',
    'US/Pacific-New': u'Pacific-New',
    'Chile/Continental': u'Continental',
    'Europe/Kirov': u'Kirov',
    'America/Campo_Grande': u'Campo Grande',
    'America/Dawson_Creek': u'Dawson Creek',
    'America/Rosario': u'Rosario',
    'Antarctica/Vostok': u'Vostok',
    'US/Aleutian': u'Aleutian',
    'Europe/Bucharest': u'Bucharest',
    'America/Porto_Velho': u'Porto Velho',
    'Europe/Monaco': u'Monaco',
    'NZ': u'NZ',
    'Africa/Ceuta': u'Ceuta',
    'Europe/Rome': u'Rome',
    'America/Winnipeg': u'Winnipeg',
    'America/Knox_IN': u'Knox IN',
    'Asia/Aqtobe': u'Aqtobe',
    'Africa/Dakar': u'Dakar',
    'America/Fortaleza': u'Fortaleza',
    'Pacific/Samoa': u'Samoa',
    'Pacific/Tarawa': u'Tarawa',
    'America/Dawson': u'Dawson',
    'Africa/Addis_Ababa': u'Addis Ababa',
    'Pacific/Efate': u'Efate',
    'Pacific/Johnston': u'Johnston',
    'GMT': u'GMT',
    'Australia/Brisbane': u'Brisbane',
    'Canada/Yukon': u'Yukon',
    'Asia/Qyzylorda': u'Qyzylorda',
    'Europe/San_Marino': u'San Marino',
    'Asia/Jerusalem': u'Jerusalem',
    'America/Barbados': u'Barbados',
    'Pacific/Auckland': u'Auckland',
    'America/Metlakatla': u'Metlakatla',
    'Etc/Zulu': u'Zulu',
    'America/Denver': u'Denver',
    'Indian/Chagos': u'Chagos',
    'America/Glace_Bay': u'Glace Bay',
    'Pacific/Bougainville': u'Bougainville Island',
    'Africa/Gaborone': u'Gaborone',
    'Africa/Tunis': u'Tunis',
    'Australia/Tasmania': u'Tasmania',
    'Asia/Ust-Nera': u'Ust-Nera',
    'Europe/Vaduz': u'Vaduz',
    'Africa/Lubumbashi': u'Lubumbashi',
    'America/Resolute': u'Resolute',
    'Etc/GMT': u'GMT',
    'Asia/Gaza': u'Gaza',
    'Europe/Dublin': u'Dublin',
    'GMT+0': u'GMT+0',
    'Antarctica/DumontDUrville': u'Dumont d’Urville',
    'America/Argentina/Catamarca': u'Catamarca',
    'America/Indiana/Knox': u'Knox, Indiana',
    'Etc/GMT-6': u'GMT-6',
    'Africa/Timbuktu': u'Timbuktu',
    'Asia/Novosibirsk': u'Novosibirsk',
    'EST': u'EST',
    'Mexico/BajaNorte': u'BajaNorte',
    'Africa/Kigali': u'Kigali',
    'Brazil/West': u'West',
    'America/Grand_Turk': u'Grand Turk',
    'Mexico/BajaSur': u'BajaSur',
    'America/Argentina/Salta': u'Salta',
    'Canada/Central': u'Central',
    'Africa/Lagos': u'Lagos',
    'GMT-0': u'GMT-0',
    'Europe/Sofia': u'Sofia',
    'Singapore': u'Singapore',
    'Africa/Tripoli': u'Tripoli',
    'America/Anchorage': u'Anchorage',
    'America/Monterrey': u'Monterrey',
    'Pacific/Nauru': u'Nauru',
}

territories = {
    'BD': u'Bangladesh',
    'BE': u'Belgium',
    'BF': u'Burkina Faso',
    'BG': u'Bulgaria',
    'VE': u'Venezuela',
    'BA': u'Bosnia & Herzegovina',
    'BB': u'Barbados',
    'WF': u'Wallis & Futuna',
    'BL': u'St. Barthélemy',
    'BM': u'Bermuda',
    'BN': u'Brunei',
    'BO': u'Bolivia',
    'BH': u'Bahrain',
    'BI': u'Burundi',
    'BJ': u'Benin',
    'BT': u'Bhutan',
    '011': u'Western Africa',
    'BV': u'Bouvet Island',
    'BW': u'Botswana',
    '014': u'Eastern Africa',
    '015': u'Northern Africa',
    'BR': u'Brazil',
    '017': u'Middle Africa',
    '018': u'Southern Africa',
    '019': u'Americas',
    'BY': u'Belarus',
    'BZ': u'Belize',
    'LV': u'Latvia',
    'RW': u'Rwanda',
    'RS': u'Serbia',
    'TL': u'Timor-Leste',
    'RE': u'Réunion',
    'TM': u'Turkmenistan',
    'OM': u'Oman',
    'TJ': u'Tajikistan',
    'RO': u'Romania',
    'TK': u'Tokelau',
    'GW': u'Guinea-Bissau',
    'GU': u'Guam',
    'GT': u'Guatemala',
    'GS': u'South Georgia & South Sandwich Islands',
    'GR': u'Greece',
    'GQ': u'Equatorial Guinea',
    'GP': u'Guadeloupe',
    'JP': u'Japan',
    'KI': u'Kiribati',
    'GY': u'Guyana',
    'GG': u'Guernsey',
    'GF': u'French Guiana',
    'GE': u'Georgia',
    'GD': u'Grenada',
    '021': u'Northern America',
    'GB': u'United Kingdom',
    'GA': u'Gabon',
    'SV': u'El Salvador',
    'GN': u'Guinea',
    'GM': u'Gambia',
    'GL': u'Greenland',
    '029': u'Caribbean',
    'GI': u'Gibraltar',
    'GH': u'Ghana',
    'JE': u'Jersey',
    'TN': u'Tunisia',
    'JM': u'Jamaica',
    '013': u'Central America',
    'WS': u'Samoa',
    'TA': u'Tristan da Cunha',
    '419': u'Latin America',
    'BQ': u'Caribbean Netherlands',
    'HR': u'Croatia',
    'BS': u'Bahamas',
    'HT': u'Haiti',
    'HU': u'Hungary',
    'HK': u'Hong Kong SAR China',
    '039': u'Southern Europe',
    'HN': u'Honduras',
    'FI': u'Finland',
    '142': u'Asia',
    '030': u'Eastern Asia',
    '034': u'Southern Asia',
    '035': u'Southeast Asia',
    'PR': u'Puerto Rico',
    'PS': u'Palestinian Territories',
    'FK': u'Falkland Islands',
    'IO': u'British Indian Ocean Territory',
    'PW': u'Palau',
    'PT': u'Portugal',
    'SJ': u'Svalbard & Jan Mayen',
    'MD': u'Moldova',
    'IQ': u'Iraq',
    'PA': u'Panama',
    'PF': u'French Polynesia',
    'PG': u'Papua New Guinea',
    'PE': u'Peru',
    'PK': u'Pakistan',
    'PH': u'Philippines',
    'FO': u'Faroe Islands',
    'PN': u'Pitcairn Islands',
    'PL': u'Poland',
    'PM': u'St. Pierre & Miquelon',
    'ZM': u'Zambia',
    'EH': u'Western Sahara',
    'RU': u'Russia',
    'EE': u'Estonia',
    'EG': u'Egypt',
    'EA': u'Ceuta & Melilla',
    'ZA': u'South Africa',
    'EC': u'Ecuador',
    'IT': u'Italy',
    'VN': u'Vietnam',
    'ZZ': u'Unknown Region',
    'SB': u'Solomon Islands',
    'EU': u'European Union',
    'ET': u'Ethiopia',
    'SO': u'Somalia',
    'ZW': u'Zimbabwe',
    'SA': u'Saudi Arabia',
    'ES': u'Spain',
    'ER': u'Eritrea',
    'ME': u'Montenegro',
    'AQ': u'Antarctica',
    'MG': u'Madagascar',
    'MF': u'St. Martin',
    'MA': u'Morocco',
    'MC': u'Monaco',
    'UZ': u'Uzbekistan',
    'MM': u'Myanmar (Burma)',
    'ML': u'Mali',
    'MO': u'Macau SAR China',
    'MN': u'Mongolia',
    'HM': u'Heard & McDonald Islands',
    'AS': u'American Samoa',
    'MK': u'Macedonia',
    'MU': u'Mauritius',
    'MT': u'Malta',
    'MW': u'Malawi',
    'MV': u'Maldives',
    'MQ': u'Martinique',
    'MP': u'Northern Mariana Islands',
    'MS': u'Montserrat',
    'MR': u'Mauritania',
    'IM': u'Isle of Man',
    'UG': u'Uganda',
    'TZ': u'Tanzania',
    'MY': u'Malaysia',
    'MX': u'Mexico',
    'IL': u'Israel',
    'IC': u'Canary Islands',
    'FR': u'France',
    'AW': u'Aruba',
    'DO': u'Dominican Republic',
    'SH': u'St. Helena',
    'VC': u'St. Vincent & Grenadines',
    '054': u'Melanesia',
    'DJ': u'Djibouti',
    'FJ': u'Fiji',
    '057': u'Micronesian Region',
    'FM': u'Micronesia',
    '053': u'Australasia',
    'NI': u'Nicaragua',
    'NL': u'Netherlands',
    'NO': u'Norway',
    'NA': u'Namibia',
    'VU': u'Vanuatu',
    'NC': u'New Caledonia',
    'NE': u'Niger',
    'NF': u'Norfolk Island',
    'NG': u'Nigeria',
    'NZ': u'New Zealand',
    'NP': u'Nepal',
    'NR': u'Nauru',
    'NU': u'Niue',
    '061': u'Polynesia',
    'XK': u'Kosovo',
    'CI': u'Côte d’Ivoire',
    'CH': u'Switzerland',
    'CO': u'Colombia',
    'CN': u'China',
    'CM': u'Cameroon',
    'CL': u'Chile',
    'CC': u'Cocos (Keeling) Islands',
    'CA': u'Canada',
    'CG': u'Congo - Brazzaville',
    'CF': u'Central African Republic',
    'CD': u'Congo - Kinshasa',
    'CZ': u'Czech Republic',
    'CY': u'Cyprus',
    'CX': u'Christmas Island',
    'CR': u'Costa Rica',
    'PY': u'Paraguay',
    'CP': u'Clipperton Island',
    'CW': u'Curaçao',
    'CV': u'Cape Verde',
    'CU': u'Cuba',
    'SZ': u'Swaziland',
    'SY': u'Syria',
    'SX': u'Sint Maarten',
    'KG': u'Kyrgyzstan',
    'KE': u'Kenya',
    'SS': u'South Sudan',
    'SR': u'Suriname',
    '143': u'Central Asia',
    'KH': u'Cambodia',
    'KN': u'St. Kitts & Nevis',
    'KM': u'Comoros',
    'ST': u'São Tomé & Príncipe',
    'SK': u'Slovakia',
    'KR': u'South Korea',
    'SI': u'Slovenia',
    'KP': u'North Korea',
    'KW': u'Kuwait',
    'SN': u'Senegal',
    'SM': u'San Marino',
    'SL': u'Sierra Leone',
    'SC': u'Seychelles',
    'KZ': u'Kazakhstan',
    'KY': u'Cayman Islands',
    'SG': u'Singapore',
    'SE': u'Sweden',
    'SD': u'Sudan',
    '151': u'Eastern Europe',
    '150': u'Europe',
    'DM': u'Dominica',
    '155': u'Western Europe',
    '154': u'Northern Europe',
    'VG': u'British Virgin Islands',
    'DG': u'Diego Garcia',
    'DE': u'Germany',
    'YE': u'Yemen',
    'MH': u'Marshall Islands',
    'DZ': u'Algeria',
    'US': u'United States',
    'UY': u'Uruguay',
    'YT': u'Mayotte',
    'UM': u'U.S. Outlying Islands',
    'LB': u'Lebanon',
    'LC': u'St. Lucia',
    'LA': u'Laos',
    'TV': u'Tuvalu',
    'TW': u'Taiwan',
    'TT': u'Trinidad & Tobago',
    'TR': u'Turkey',
    'LK': u'Sri Lanka',
    'LI': u'Liechtenstein',
    'CK': u'Cook Islands',
    'TO': u'Tonga',
    'LT': u'Lithuania',
    'LU': u'Luxembourg',
    'LR': u'Liberia',
    'LS': u'Lesotho',
    'TH': u'Thailand',
    'TF': u'French Southern Territories',
    'TG': u'Togo',
    'TD': u'Chad',
    'TC': u'Turks & Caicos Islands',
    'LY': u'Libya',
    'VA': u'Vatican City',
    'AC': u'Ascension Island',
    '145': u'Western Asia',
    'AE': u'United Arab Emirates',
    'AD': u'Andorra',
    'AG': u'Antigua & Barbuda',
    'AF': u'Afghanistan',
    'AI': u'Anguilla',
    'VI': u'U.S. Virgin Islands',
    'IS': u'Iceland',
    'IR': u'Iran',
    'AM': u'Armenia',
    'AL': u'Albania',
    'AO': u'Angola',
    '003': u'North America',
    '002': u'Africa',
    '001': u'World',
    'AR': u'Argentina',
    'AU': u'Australia',
    'AT': u'Austria',
    '005': u'South America',
    'IN': u'India',
    'AX': u'Åland Islands',
    '009': u'Oceania',
    'AZ': u'Azerbaijan',
    'IE': u'Ireland',
    'ID': u'Indonesia',
    'JO': u'Jordan',
    'UA': u'Ukraine',
    'QA': u'Qatar',
    'DK': u'Denmark',
    'MZ': u'Mozambique',
    'QO': u'Outlying Oceania',
}


import six

from .translation import L18NMap, L18NListMap

try:
    from . import __maps
    tz_cities = L18NMap(__maps.tz_cities)
    territories = L18NMap(__maps.territories)

    # tz_fullnames requires a main dictionary and an auxiliary translations
    # dictionary (for components)

    _main_dict = dict(__maps.tz_cities)
    _aux_dict = {}
    for k, v in six.iteritems(__maps.tz_locations):
        if k in _main_dict:
            _main_dict[k] = v
        else:
            _aux_dict[k] = v

    tz_fullnames = L18NListMap('/', _aux_dict, _main_dict)

except ImportError:
    tz_cities = {}
    tz_fullnames = {}
    territories = {}



import os
import gettext
import bisect
from locale import getdefaultlocale
from collections.abc import MutableMapping
from copy import copy, deepcopy

import six


class Trans:

    def __init__(self):
        self.registry = {}
        self.current = None
        self.set(getdefaultlocale()[0])

    def __getitem__(self, language):
        if language:
            try:
                return self.registry[language]
            except KeyError:
                self.registry[language] = gettext.translation(
                    'l18n',
                    os.path.join(os.path.dirname(__file__), 'locale'),
                    languages=[language],
                    fallback=True
                )
                return self.registry[language]
        else:
            return None

    def set(self, language):
        self.current = self[language]

    def gettext(self, s):
        try:
            return self.current.gettext(s)
        except AttributeError:
            return s

    if six.PY2:
        def ugettext(self, s):
            try:
                return self.current.ugettext(s)
            except AttributeError:
                return s


_trans = Trans()


def set_language(language=None):
    _trans.set(language)


if six.PY2:
    def translate(s, utf8=True, trans=_trans):
        if trans:
            if utf8:
                return trans.ugettext(s)
            return trans.gettext(s)
        else:
            return s
else:
    def translate(s, utf8=True, trans=_trans):
        if trans:
            t = trans.gettext(s)
            if utf8:
                return t
            return t.encode()
        else:
            return s


class L18NLazyObject:

    def _value(self, utf8=True):
        raise NotImplementedError

    def __str__(self):
        return self._value(utf8=six.PY3)

    def __bytes__(self):
        return self._value(utf8=False)

    def __unicode__(self):
        return self._value(utf8=True)


class L18NLazyString(L18NLazyObject):

    def __init__(self, s):
        self._str = s

    def __copy__(self):
        return self.__class__(self._str)

    def __deepcopy__(self, memo):
        result = self.__copy__()
        memo[id(self)] = result
        return result

    def _value(self, utf8=True):
        return translate(self._str, utf8)

    def __repr__(self):
        return 'L18NLazyString <%s>' % repr(self._str)

    def __getattr__(self, name):
        # fallback to call the value's attribute in case it's not found in
        # L18NLazyString
        return getattr(self._value(), name)


class L18NLazyStringsList(L18NLazyObject):

    def __init__(self, sep='/', *s):
        # we assume that the separator and the strings have the same encoding
        # (text_type)
        self._sep = sep
        self._strings = s

    def __copy__(self):
        return self.__class__(self._sep, *self._strings)

    def __deepcopy__(self, memo):
        result = self.__copy__()
        memo[id(self)] = result
        return result

    def _value(self, utf8=True):
        sep = self._sep
        if utf8 and isinstance(sep, six.binary_type):
            sep = sep.decode(encoding='utf-8')
        elif not utf8 and isinstance(sep, six.text_type):
            sep = sep.encode(encoding='utf-8')
        return sep.join([translate(s, utf8)
                         for s in self._strings])

    def __repr__(self):
        return 'L18NLazyStringsList <%s>' % self._sep.join([
            repr(s) for s in self._strings
        ])

    def __getattr__(self, name):
        # fallback to call the value's attribute in case it's not found in
        # L18NLazyStringsList
        return getattr(self._value(), name)


class L18NBaseMap(MutableMapping):
    """
    Generic dictionary that returns lazy string or lazy string lists
    """

    def __init__(self, *args, **kwargs):
        self.store = dict(*args, **kwargs)
        self.sorted = {}

    def __copy__(self):
        result = self.__class__()
        result.store = self.store
        result.sorted = self.sorted
        return result

    def __deepcopy__(self, memo):
        result = self.__class__()
        memo[id(self)] = result
        result.store = deepcopy(self.store, memo)
        result.sorted = deepcopy(self.sorted, memo)
        return result

    def __getitem__(self, key):
        raise NotImplementedError

    def __setitem__(self, key, value):
        self.store[key] = value
        for locale, (keys, values) in six.iteritems(self.sorted):
            tr = translate(value, trans=_trans[locale])
            i = bisect.bisect_left(values, tr)
            keys.insert(i, key)
            values.insert(i, tr)

    def __delitem__(self, key):
        del self.store[key]
        for keys, values in self.sorted.values():
            i = keys.index(key)
            del keys[i]
            del values[i]

    def __iter__(self):
        loc = _trans.current._info['language'] if _trans.current else None
        try:
            return iter(self.sorted[loc][0])
        except KeyError:
            keys = []
            values = []
            # we can't use iteritems here, as we need to call __getitem__
            # via self[key]
            for key in iter(self.store):
                value = six.text_type(self[key])
                i = bisect.bisect_left(values, value)
                keys.insert(i, key)
                values.insert(i, value)
            self.sorted[loc] = (keys, values)
            return iter(keys)

    def __len__(self):
        return len(self.store)

    def subset(self, keys):
        """
        Generates a subset of the current map (e.g. to retrieve only tzs in
        common_timezones from the tz_cities or tz_fullnames maps)
        """
        sub = self.__class__()

        self_keys = set(self.store.keys())
        subset_keys = self_keys.intersection(keys)
        removed_keys = self_keys.difference(subset_keys)

        sub.store = {k: self.store[k] for k in subset_keys}
        for loc, sorted_items in six.iteritems(self.sorted):
            loc_keys = copy(self.sorted[loc][0])
            loc_values = copy(self.sorted[loc][1])
            for k in removed_keys:
                i = loc_keys.index(k)
                del loc_keys[i]
                del loc_values[i]
            sub.sorted[loc] = (loc_keys, loc_values)
        return sub


class L18NMap(L18NBaseMap):

    def __getitem__(self, key):
        return L18NLazyString(self.store[key])


class L18NListMap(L18NBaseMap):

    def __init__(self, sep='/', aux=None, *args, **kwargs):
        self._sep = sep
        self._aux = aux
        super(L18NListMap, self).__init__(*args, **kwargs)

    def __copy__(self):
        result = super(L18NListMap, self).__copy__()
        result._sep = self._sep
        result._aux = self._aux
        return result

    def __deepcopy__(self, memo):
        result = super(L18NListMap, self).__deepcopy__(memo)
        result._sep = self._sep
        result._aux = None if self._aux is None else deepcopy(self._aux, memo)
        return result

    def __getitem__(self, key):
        strs = key.split(self._sep)
        strs[-1] = key
        lst = []
        for s in strs:
            try:
                lst.append(self.store[s])
            except KeyError:
                lst.append(self._aux[s])
        return L18NLazyStringsList(self._sep, *lst)

    def subset(self, keys):
        sub = super(L18NListMap, self).subset(keys)
        sub._sep = self._sep
        sub._aux = deepcopy(self._aux)
        return sub



from time import time

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import PermissionDenied
from django.core.paginator import Paginator
from django.db import transaction
from django.db.models import Count
from django.http import Http404, HttpResponse, JsonResponse
from django.http.request import QueryDict
from django.shortcuts import get_object_or_404, redirect, render
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils import timezone
from django.utils.http import is_safe_url, urlquote
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext as _
from django.views.decorators.http import require_GET, require_POST
from django.views.decorators.vary import vary_on_headers
from django.views.generic import View

from wagtail.admin import messages, signals
from wagtail.admin.action_menu import PageActionMenu
from wagtail.admin.auth import user_has_any_page_permission, user_passes_test
from wagtail.admin.forms.pages import CopyForm
from wagtail.admin.forms.search import SearchForm
from wagtail.admin.mail import send_notification
from wagtail.admin.navigation import get_explorable_root_page
from wagtail.core import hooks
from wagtail.core.models import Page, PageRevision, UserPagePermissionsProxy
from wagtail.search.query import MATCH_ALL


def get_valid_next_url_from_request(request):
    next_url = request.POST.get('next') or request.GET.get('next')
    if not next_url or not is_safe_url(url=next_url, allowed_hosts={request.get_host()}):
        return ''
    return next_url


@user_passes_test(user_has_any_page_permission)
def index(request, parent_page_id=None):
    if parent_page_id:
        parent_page = get_object_or_404(Page, id=parent_page_id)
    else:
        parent_page = Page.get_first_root_node()

    # This will always succeed because of the @user_passes_test above.
    root_page = get_explorable_root_page(request.user)

    # If this page isn't a descendant of the user's explorable root page,
    # then redirect to that explorable root page instead.
    if not (
        parent_page.pk == root_page.pk
        or parent_page.is_descendant_of(root_page)
    ):
        return redirect('wagtailadmin_explore', root_page.pk)

    parent_page = parent_page.specific

    user_perms = UserPagePermissionsProxy(request.user)
    pages = (
        parent_page.get_children().prefetch_related(
            "content_type", "sites_rooted_here"
        )
        & user_perms.explorable_pages()
    )

    # Get page ordering
    ordering = request.GET.get('ordering', '-latest_revision_created_at')
    if ordering not in [
        'title',
        '-title',
        'content_type',
        '-content_type',
        'live', '-live',
        'latest_revision_created_at',
        '-latest_revision_created_at',
        'ord'
    ]:
        ordering = '-latest_revision_created_at'

    if ordering == 'ord':
        # preserve the native ordering from get_children()
        pass
    elif ordering == 'latest_revision_created_at':
        # order by oldest revision first.
        # Special case NULL entries - these should go at the top of the list.
        # Do this by annotating with Count('latest_revision_created_at'),
        # which returns 0 for these
        pages = pages.annotate(
            null_position=Count('latest_revision_created_at')
        ).order_by('null_position', 'latest_revision_created_at')
    elif ordering == '-latest_revision_created_at':
        # order by oldest revision first.
        # Special case NULL entries - these should go at the end of the list.
        pages = pages.annotate(
            null_position=Count('latest_revision_created_at')
        ).order_by('-null_position', '-latest_revision_created_at')
    else:
        pages = pages.order_by(ordering)

    # Don't paginate if sorting by page order - all pages must be shown to
    # allow drag-and-drop reordering
    do_paginate = ordering != 'ord'

    if do_paginate or pages.count() < 100:
        # Retrieve pages in their most specific form, so that custom
        # get_admin_display_title and get_url_parts methods on subclasses are respected.
        # However, skip this on unpaginated listings with >100 child pages as this could
        # be a significant performance hit. (This should only happen on the reorder view,
        # and hopefully no-one is having to do manual reordering on listings that large...)
        pages = pages.specific(defer=True)

    # allow hooks to modify the queryset
    for hook in hooks.get_hooks('construct_explorer_page_queryset'):
        pages = hook(parent_page, pages, request)

    # Pagination
    if do_paginate:
        paginator = Paginator(pages, per_page=50)
        pages = paginator.get_page(request.GET.get('p'))

    return render(request, 'wagtailadmin/pages/index.html', {
        'parent_page': parent_page.specific,
        'ordering': ordering,
        'pagination_query_params': "ordering=%s" % ordering,
        'pages': pages,
        'do_paginate': do_paginate,
    })


def add_subpage(request, parent_page_id):
    parent_page = get_object_or_404(Page, id=parent_page_id).specific
    if not parent_page.permissions_for_user(request.user).can_add_subpage():
        raise PermissionDenied

    page_types = [
        (model.get_verbose_name(), model._meta.app_label, model._meta.model_name)
        for model in type(parent_page).creatable_subpage_models()
        if model.can_create_at(parent_page)
    ]
    # sort by lower-cased version of verbose name
    page_types.sort(key=lambda page_type: page_type[0].lower())

    if len(page_types) == 1:
        # Only one page type is available - redirect straight to the create form rather than
        # making the user choose
        verbose_name, app_label, model_name = page_types[0]
        return redirect('wagtailadmin_pages:add', app_label, model_name, parent_page.id)

    return render(request, 'wagtailadmin/pages/add_subpage.html', {
        'parent_page': parent_page,
        'page_types': page_types,
        'next': get_valid_next_url_from_request(request),
    })


def content_type_use(request, content_type_app_name, content_type_model_name):
    try:
        content_type = ContentType.objects.get_by_natural_key(content_type_app_name, content_type_model_name)
    except ContentType.DoesNotExist:
        raise Http404

    page_class = content_type.model_class()

    # page_class must be a Page type and not some other random model
    if not issubclass(page_class, Page):
        raise Http404

    pages = page_class.objects.all()

    paginator = Paginator(pages, per_page=10)
    pages = paginator.get_page(request.GET.get('p'))

    return render(request, 'wagtailadmin/pages/content_type_use.html', {
        'pages': pages,
        'app_name': content_type_app_name,
        'content_type': content_type,
        'page_class': page_class,
    })


def create(request, content_type_app_name, content_type_model_name, parent_page_id):
    parent_page = get_object_or_404(Page, id=parent_page_id).specific
    parent_page_perms = parent_page.permissions_for_user(request.user)
    if not parent_page_perms.can_add_subpage():
        raise PermissionDenied

    try:
        content_type = ContentType.objects.get_by_natural_key(content_type_app_name, content_type_model_name)
    except ContentType.DoesNotExist:
        raise Http404

    # Get class
    page_class = content_type.model_class()

    # Make sure the class is a descendant of Page
    if not issubclass(page_class, Page):
        raise Http404

    # page must be in the list of allowed subpage types for this parent ID
    if page_class not in parent_page.creatable_subpage_models():
        raise PermissionDenied

    if not page_class.can_create_at(parent_page):
        raise PermissionDenied

    for fn in hooks.get_hooks('before_create_page'):
        result = fn(request, parent_page, page_class)
        if hasattr(result, 'status_code'):
            return result

    page = page_class(owner=request.user)
    edit_handler = page_class.get_edit_handler()
    edit_handler = edit_handler.bind_to(request=request, instance=page)
    form_class = edit_handler.get_form_class()

    next_url = get_valid_next_url_from_request(request)

    if request.method == 'POST':
        form = form_class(request.POST, request.FILES, instance=page,
                          parent_page=parent_page)

        if form.is_valid():
            page = form.save(commit=False)

            is_publishing = bool(request.POST.get('action-publish')) and parent_page_perms.can_publish_subpage()
            is_submitting = bool(request.POST.get('action-submit'))

            if not is_publishing:
                page.live = False

            # Save page
            parent_page.add_child(instance=page)

            # Save revision
            revision = page.save_revision(
                user=request.user,
                submitted_for_moderation=is_submitting,
            )

            # Publish
            if is_publishing:
                revision.publish()

            # Notifications
            if is_publishing:
                if page.go_live_at and page.go_live_at > timezone.now():
                    messages.success(request, _("Page '{0}' created and scheduled for publishing.").format(page.get_admin_display_title()), buttons=[
                        messages.button(reverse('wagtailadmin_pages:edit', args=(page.id,)), _('Edit'))
                    ])
                else:
                    buttons = []
                    if page.url is not None:
                        buttons.append(messages.button(page.url, _('View live'), new_window=True))
                    buttons.append(messages.button(reverse('wagtailadmin_pages:edit', args=(page.id,)), _('Edit')))
                    messages.success(request, _("Page '{0}' created and published.").format(page.get_admin_display_title()), buttons=buttons)
            elif is_submitting:
                messages.success(
                    request,
                    _("Page '{0}' created and submitted for moderation.").format(page.get_admin_display_title()),
                    buttons=[
                        messages.button(
                            reverse('wagtailadmin_pages:view_draft', args=(page.id,)),
                            _('View draft'),
                            new_window=True
                        ),
                        messages.button(
                            reverse('wagtailadmin_pages:edit', args=(page.id,)),
                            _('Edit')
                        )
                    ]
                )
                if not send_notification(page.get_latest_revision().id, 'submitted', request.user.pk):
                    messages.error(request, _("Failed to send notifications to moderators"))
            else:
                messages.success(request, _("Page '{0}' created.").format(page.get_admin_display_title()))

            for fn in hooks.get_hooks('after_create_page'):
                result = fn(request, page)
                if hasattr(result, 'status_code'):
                    return result

            if is_publishing or is_submitting:
                # we're done here
                if next_url:
                    # redirect back to 'next' url if present
                    return redirect(next_url)
                # redirect back to the explorer
                return redirect('wagtailadmin_explore', page.get_parent().id)
            else:
                # Just saving - remain on edit page for further edits
                target_url = reverse('wagtailadmin_pages:edit', args=[page.id])
                if next_url:
                    # Ensure the 'next' url is passed through again if present
                    target_url += '?next=%s' % urlquote(next_url)
                return redirect(target_url)
        else:
            messages.validation_error(
                request, _("The page could not be created due to validation errors"), form
            )
            has_unsaved_changes = True
    else:
        signals.init_new_page.send(sender=create, page=page, parent=parent_page)
        form = form_class(instance=page, parent_page=parent_page)
        has_unsaved_changes = False

    edit_handler = edit_handler.bind_to(form=form)

    return render(request, 'wagtailadmin/pages/create.html', {
        'content_type': content_type,
        'page_class': page_class,
        'parent_page': parent_page,
        'edit_handler': edit_handler,
        'action_menu': PageActionMenu(request, view='create', parent_page=parent_page),
        'preview_modes': page.preview_modes,
        'form': form,
        'next': next_url,
        'has_unsaved_changes': has_unsaved_changes,
    })


def edit(request, page_id):
    real_page_record = get_object_or_404(Page, id=page_id)
    latest_revision = real_page_record.get_latest_revision()
    page = real_page_record.get_latest_revision_as_page()
    parent = page.get_parent()

    content_type = ContentType.objects.get_for_model(page)
    page_class = content_type.model_class()

    page_perms = page.permissions_for_user(request.user)
    if not page_perms.can_edit():
        raise PermissionDenied

    for fn in hooks.get_hooks('before_edit_page'):
        result = fn(request, page)
        if hasattr(result, 'status_code'):
            return result

    edit_handler = page_class.get_edit_handler()
    edit_handler = edit_handler.bind_to(instance=page, request=request)
    form_class = edit_handler.get_form_class()

    next_url = get_valid_next_url_from_request(request)

    errors_debug = None

    if request.method == 'POST':
        form = form_class(request.POST, request.FILES, instance=page,
                          parent_page=parent)

        if form.is_valid() and not page.locked:
            page = form.save(commit=False)

            is_publishing = bool(request.POST.get('action-publish')) and page_perms.can_publish()
            is_submitting = bool(request.POST.get('action-submit'))
            is_reverting = bool(request.POST.get('revision'))

            # If a revision ID was passed in the form, get that revision so its
            # date can be referenced in notification messages
            if is_reverting:
                previous_revision = get_object_or_404(page.revisions, id=request.POST.get('revision'))

            # Save revision
            revision = page.save_revision(
                user=request.user,
                submitted_for_moderation=is_submitting,
            )
            # store submitted go_live_at for messaging below
            go_live_at = page.go_live_at

            # Publish
            if is_publishing:
                revision.publish()
                # Need to reload the page because the URL may have changed, and we
                # need the up-to-date URL for the "View Live" button.
                page = page.specific_class.objects.get(pk=page.pk)

            # Notifications
            if is_publishing:
                if go_live_at and go_live_at > timezone.now():
                    # Page has been scheduled for publishing in the future

                    if is_reverting:
                        message = _(
                            "Revision from {0} of page '{1}' has been scheduled for publishing."
                        ).format(
                            previous_revision.created_at.strftime("%d %b %Y %H:%M"),
                            page.get_admin_display_title()
                        )
                    else:
                        if page.live:
                            message = _(
                                "Page '{0}' is live and this revision has been scheduled for publishing."
                            ).format(
                                page.get_admin_display_title()
                            )
                        else:
                            message = _(
                                "Page '{0}' has been scheduled for publishing."
                            ).format(
                                page.get_admin_display_title()
                            )

                    messages.success(request, message, buttons=[
                        messages.button(
                            reverse('wagtailadmin_pages:edit', args=(page.id,)),
                            _('Edit')
                        )
                    ])

                else:
                    # Page is being published now

                    if is_reverting:
                        message = _(
                            "Revision from {0} of page '{1}' has been published."
                        ).format(
                            previous_revision.created_at.strftime("%d %b %Y %H:%M"),
                            page.get_admin_display_title()
                        )
                    else:
                        message = _(
                            "Page '{0}' has been published."
                        ).format(
                            page.get_admin_display_title()
                        )

                    buttons = []
                    if page.url is not None:
                        buttons.append(messages.button(page.url, _('View live'), new_window=True))
                    buttons.append(messages.button(reverse('wagtailadmin_pages:edit', args=(page_id,)), _('Edit')))
                    messages.success(request, message, buttons=buttons)

            elif is_submitting:

                message = _(
                    "Page '{0}' has been submitted for moderation."
                ).format(
                    page.get_admin_display_title()
                )

                messages.success(request, message, buttons=[
                    messages.button(
                        reverse('wagtailadmin_pages:view_draft', args=(page_id,)),
                        _('View draft'),
                        new_window=True
                    ),
                    messages.button(
                        reverse('wagtailadmin_pages:edit', args=(page_id,)),
                        _('Edit')
                    )
                ])

                if not send_notification(page.get_latest_revision().id, 'submitted', request.user.pk):
                    messages.error(request, _("Failed to send notifications to moderators"))

            else:  # Saving

                if is_reverting:
                    message = _(
                        "Page '{0}' has been replaced with revision from {1}."
                    ).format(
                        page.get_admin_display_title(),
                        previous_revision.created_at.strftime("%d %b %Y %H:%M")
                    )
                else:
                    message = _(
                        "Page '{0}' has been updated."
                    ).format(
                        page.get_admin_display_title()
                    )

                messages.success(request, message)

            for fn in hooks.get_hooks('after_edit_page'):
                result = fn(request, page)
                if hasattr(result, 'status_code'):
                    return result

            if is_publishing or is_submitting:
                # we're done here - redirect back to the explorer
                if next_url:
                    # redirect back to 'next' url if present
                    return redirect(next_url)
                # redirect back to the explorer
                return redirect('wagtailadmin_explore', page.get_parent().id)
            else:
                # Just saving - remain on edit page for further edits
                target_url = reverse('wagtailadmin_pages:edit', args=[page.id])
                if next_url:
                    # Ensure the 'next' url is passed through again if present
                    target_url += '?next=%s' % urlquote(next_url)
                return redirect(target_url)
        else:
            if page.locked:
                messages.error(request, _("The page could not be saved as it is locked"))
            else:
                messages.validation_error(
                    request, _("The page could not be saved due to validation errors"), form
                )
            errors_debug = (
                repr(form.errors)
                + repr([
                    (name, formset.errors)
                    for (name, formset) in form.formsets.items()
                    if formset.errors
                ])
            )
            has_unsaved_changes = True
    else:
        form = form_class(instance=page, parent_page=parent)
        has_unsaved_changes = False

    edit_handler = edit_handler.bind_to(form=form)

    # Check for revisions still undergoing moderation and warn
    if latest_revision and latest_revision.submitted_for_moderation:
        buttons = []

        if page.live:
            buttons.append(messages.button(
                reverse('wagtailadmin_pages:revisions_compare', args=(page.id, 'live', latest_revision.id)),
                _('Compare with live version')
            ))

        messages.warning(request, _("This page is currently awaiting moderation"), buttons=buttons)

    if page.live and page.has_unpublished_changes:
        # Page status needs to present the version of the page containing the correct live URL
        page_for_status = real_page_record.specific
    else:
        page_for_status = page

    return render(request, 'wagtailadmin/pages/edit.html', {
        'page': page,
        'page_for_status': page_for_status,
        'content_type': content_type,
        'edit_handler': edit_handler,
        'errors_debug': errors_debug,
        'action_menu': PageActionMenu(request, view='edit', page=page),
        'preview_modes': page.preview_modes,
        'form': form,
        'next': next_url,
        'has_unsaved_changes': has_unsaved_changes,
    })


def delete(request, page_id):
    page = get_object_or_404(Page, id=page_id).specific
    if not page.permissions_for_user(request.user).can_delete():
        raise PermissionDenied

    with transaction.atomic():
        for fn in hooks.get_hooks('before_delete_page'):
            result = fn(request, page)
            if hasattr(result, 'status_code'):
                return result

        next_url = get_valid_next_url_from_request(request)

        if request.method == 'POST':
            parent_id = page.get_parent().id
            page.delete()

            messages.success(request, _("Page '{0}' deleted.").format(page.get_admin_display_title()))

            for fn in hooks.get_hooks('after_delete_page'):
                result = fn(request, page)
                if hasattr(result, 'status_code'):
                    return result

            if next_url:
                return redirect(next_url)
            return redirect('wagtailadmin_explore', parent_id)

    return render(request, 'wagtailadmin/pages/confirm_delete.html', {
        'page': page,
        'descendant_count': page.get_descendant_count(),
        'next': next_url,
    })


def view_draft(request, page_id):
    page = get_object_or_404(Page, id=page_id).get_latest_revision_as_page()
    perms = page.permissions_for_user(request.user)
    if not (perms.can_publish() or perms.can_edit()):
        raise PermissionDenied
    return page.make_preview_request(request, page.default_preview_mode)


class PreviewOnEdit(View):
    http_method_names = ('post', 'get')
    preview_expiration_timeout = 60 * 60 * 24  # seconds
    session_key_prefix = 'wagtail-preview-'

    def remove_old_preview_data(self):
        expiration = time() - self.preview_expiration_timeout
        expired_keys = [
            k for k, v in self.request.session.items()
            if k.startswith(self.session_key_prefix) and v[1] < expiration]
        # Removes the session key gracefully
        for k in expired_keys:
            self.request.session.pop(k)

    @property
    def session_key(self):
        return self.session_key_prefix + ','.join(self.args)

    def get_page(self):
        return get_object_or_404(Page,
                                 id=self.args[0]).get_latest_revision_as_page()

    def get_form(self, page, query_dict):
        form_class = page.get_edit_handler().get_form_class()
        parent_page = page.get_parent().specific

        if self.session_key not in self.request.session:
            # Session key not in session, returning null form
            return form_class(instance=page, parent_page=parent_page)

        return form_class(query_dict, instance=page, parent_page=parent_page)

    def post(self, request, *args, **kwargs):
        # TODO: Handle request.FILES.
        request.session[self.session_key] = request.POST.urlencode(), time()
        self.remove_old_preview_data()
        form = self.get_form(self.get_page(), request.POST)
        return JsonResponse({'is_valid': form.is_valid()})

    def error_response(self, page):
        return render(self.request, 'wagtailadmin/pages/preview_error.html',
                      {'page': page})

    def get(self, request, *args, **kwargs):
        page = self.get_page()

        post_data, timestamp = self.request.session.get(self.session_key,
                                                        (None, None))
        if not isinstance(post_data, str):
            post_data = ''
        form = self.get_form(page, QueryDict(post_data))

        if not form.is_valid():
            return self.error_response(page)

        form.save(commit=False)
        preview_mode = request.GET.get('mode', page.default_preview_mode)
        return page.make_preview_request(request, preview_mode)


class PreviewOnCreate(PreviewOnEdit):
    def get_page(self):
        (content_type_app_name, content_type_model_name,
         parent_page_id) = self.args
        try:
            content_type = ContentType.objects.get_by_natural_key(
                content_type_app_name, content_type_model_name)
        except ContentType.DoesNotExist:
            raise Http404

        page = content_type.model_class()()
        parent_page = get_object_or_404(Page, id=parent_page_id).specific
        # We need to populate treebeard's path / depth fields in order to
        # pass validation. We can't make these 100% consistent with the rest
        # of the tree without making actual database changes (such as
        # incrementing the parent's numchild field), but by calling treebeard's
        # internal _get_path method, we can set a 'realistic' value that will
        # hopefully enable tree traversal operations
        # to at least partially work.
        page.depth = parent_page.depth + 1
        # Puts the page at the maximum possible path
        # for a child of `parent_page`.
        page.path = Page._get_children_path_interval(parent_page.path)[1]
        return page

    def get_form(self, page, query_dict):
        form = super().get_form(page, query_dict)
        if form.is_valid():
            # Ensures our unsaved page has a suitable url.
            form.instance.set_url_path(form.parent_page)

            form.instance.full_clean()
        return form


def unpublish(request, page_id):
    page = get_object_or_404(Page, id=page_id).specific

    user_perms = UserPagePermissionsProxy(request.user)
    if not user_perms.for_page(page).can_unpublish():
        raise PermissionDenied

    next_url = get_valid_next_url_from_request(request)

    if request.method == 'POST':
        include_descendants = request.POST.get("include_descendants", False)

        page.unpublish()

        if include_descendants:
            live_descendant_pages = page.get_descendants().live().specific()
            for live_descendant_page in live_descendant_pages:
                if user_perms.for_page(live_descendant_page).can_unpublish():
                    live_descendant_page.unpublish()

        messages.success(request, _("Page '{0}' unpublished.").format(page.get_admin_display_title()), buttons=[
            messages.button(reverse('wagtailadmin_pages:edit', args=(page.id,)), _('Edit'))
        ])

        if next_url:
            return redirect(next_url)
        return redirect('wagtailadmin_explore', page.get_parent().id)

    return render(request, 'wagtailadmin/pages/confirm_unpublish.html', {
        'page': page,
        'next': next_url,
        'live_descendant_count': page.get_descendants().live().count(),
    })


def move_choose_destination(request, page_to_move_id, viewed_page_id=None):
    page_to_move = get_object_or_404(Page, id=page_to_move_id)
    page_perms = page_to_move.permissions_for_user(request.user)
    if not page_perms.can_move():
        raise PermissionDenied

    if viewed_page_id:
        viewed_page = get_object_or_404(Page, id=viewed_page_id)
    else:
        viewed_page = Page.get_first_root_node()

    viewed_page.can_choose = page_perms.can_move_to(viewed_page)

    child_pages = []
    for target in viewed_page.get_children():
        # can't move the page into itself or its descendants
        target.can_choose = page_perms.can_move_to(target)

        target.can_descend = (
            not(target == page_to_move
                or target.is_child_of(page_to_move))
            and target.get_children_count()
        )

        child_pages.append(target)

    # Pagination
    paginator = Paginator(child_pages, per_page=50)
    child_pages = paginator.get_page(request.GET.get('p'))

    return render(request, 'wagtailadmin/pages/move_choose_destination.html', {
        'page_to_move': page_to_move,
        'viewed_page': viewed_page,
        'child_pages': child_pages,
    })


def move_confirm(request, page_to_move_id, destination_id):
    page_to_move = get_object_or_404(Page, id=page_to_move_id).specific
    destination = get_object_or_404(Page, id=destination_id)
    if not page_to_move.permissions_for_user(request.user).can_move_to(destination):
        raise PermissionDenied

    if not Page._slug_is_available(page_to_move.slug, destination, page=page_to_move):
        messages.error(
            request,
            _("The slug '{0}' is already in use at the selected parent page. Make sure the slug is unique and try again".format(page_to_move.slug))
        )
        return redirect('wagtailadmin_pages:move_choose_destination', page_to_move.id, destination.id)

    for fn in hooks.get_hooks('before_move_page'):
        result = fn(request, page_to_move, destination)
        if hasattr(result, 'status_code'):
            return result

    if request.method == 'POST':
        # any invalid moves *should* be caught by the permission check above,
        # so don't bother to catch InvalidMoveToDescendant
        page_to_move.move(destination, pos='last-child')

        messages.success(request, _("Page '{0}' moved.").format(page_to_move.get_admin_display_title()), buttons=[
            messages.button(reverse('wagtailadmin_pages:edit', args=(page_to_move.id,)), _('Edit'))
        ])

        for fn in hooks.get_hooks('after_move_page'):
            result = fn(request, page_to_move)
            if hasattr(result, 'status_code'):
                return result

        return redirect('wagtailadmin_explore', destination.id)

    return render(request, 'wagtailadmin/pages/confirm_move.html', {
        'page_to_move': page_to_move,
        'destination': destination,
    })


def set_page_position(request, page_to_move_id):
    page_to_move = get_object_or_404(Page, id=page_to_move_id)
    parent_page = page_to_move.get_parent()

    if not parent_page.permissions_for_user(request.user).can_reorder_children():
        raise PermissionDenied

    if request.method == 'POST':
        # Get position parameter
        position = request.GET.get('position', None)

        # Find page thats already in this position
        position_page = None
        if position is not None:
            try:
                position_page = parent_page.get_children()[int(position)]
            except IndexError:
                pass  # No page in this position

        # Move page

        # any invalid moves *should* be caught by the permission check above,
        # so don't bother to catch InvalidMoveToDescendant

        if position_page:
            # If the page has been moved to the right, insert it to the
            # right. If left, then left.
            old_position = list(parent_page.get_children()).index(page_to_move)
            if int(position) < old_position:
                page_to_move.move(position_page, pos='left')
            elif int(position) > old_position:
                page_to_move.move(position_page, pos='right')
        else:
            # Move page to end
            page_to_move.move(parent_page, pos='last-child')

    return HttpResponse('')


@user_passes_test(user_has_any_page_permission)
def copy(request, page_id):
    page = Page.objects.get(id=page_id)

    # Parent page defaults to parent of source page
    parent_page = page.get_parent()

    # Check if the user has permission to publish subpages on the parent
    can_publish = parent_page.permissions_for_user(request.user).can_publish_subpage()

    # Create the form
    form = CopyForm(request.POST or None, user=request.user, page=page, can_publish=can_publish)

    next_url = get_valid_next_url_from_request(request)

    for fn in hooks.get_hooks('before_copy_page'):
        result = fn(request, page)
        if hasattr(result, 'status_code'):
            return result

    # Check if user is submitting
    if request.method == 'POST':
        # Prefill parent_page in case the form is invalid (as prepopulated value for the form field,
        # because ModelChoiceField seems to not fall back to the user given value)
        parent_page = Page.objects.get(id=request.POST['new_parent_page'])

        if form.is_valid():
            # Receive the parent page (this should never be empty)
            if form.cleaned_data['new_parent_page']:
                parent_page = form.cleaned_data['new_parent_page']

            if not page.permissions_for_user(request.user).can_copy_to(parent_page,
                                                                       form.cleaned_data.get('copy_subpages')):
                raise PermissionDenied

            # Re-check if the user has permission to publish subpages on the new parent
            can_publish = parent_page.permissions_for_user(request.user).can_publish_subpage()

            # Copy the page
            new_page = page.specific.copy(
                recursive=form.cleaned_data.get('copy_subpages'),
                to=parent_page,
                update_attrs={
                    'title': form.cleaned_data['new_title'],
                    'slug': form.cleaned_data['new_slug'],
                },
                keep_live=(can_publish and form.cleaned_data.get('publish_copies')),
                user=request.user,
            )

            # Give a success message back to the user
            if form.cleaned_data.get('copy_subpages'):
                messages.success(
                    request,
                    _("Page '{0}' and {1} subpages copied.").format(page.get_admin_display_title(), new_page.get_descendants().count())
                )
            else:
                messages.success(request, _("Page '{0}' copied.").format(page.get_admin_display_title()))

            for fn in hooks.get_hooks('after_copy_page'):
                result = fn(request, page, new_page)
                if hasattr(result, 'status_code'):
                    return result

            # Redirect to explore of parent page
            if next_url:
                return redirect(next_url)
            return redirect('wagtailadmin_explore', parent_page.id)

    return render(request, 'wagtailadmin/pages/copy.html', {
        'page': page,
        'form': form,
        'next': next_url,
    })


@vary_on_headers('X-Requested-With')
@user_passes_test(user_has_any_page_permission)
def search(request):
    pages = all_pages = Page.objects.all().prefetch_related('content_type').specific()
    q = MATCH_ALL
    content_types = []
    pagination_query_params = QueryDict({}, mutable=True)
    ordering = None

    if 'ordering' in request.GET:
        if request.GET['ordering'] in ['title', '-title', 'latest_revision_created_at', '-latest_revision_created_at', 'live', '-live']:
            ordering = request.GET['ordering']

            if ordering == 'title':
                pages = pages.order_by('title')
            elif ordering == '-title':
                pages = pages.order_by('-title')

            if ordering == 'latest_revision_created_at':
                pages = pages.order_by('latest_revision_created_at')
            elif ordering == '-latest_revision_created_at':
                pages = pages.order_by('-latest_revision_created_at')

            if ordering == 'live':
                pages = pages.order_by('live')
            elif ordering == '-live':
                pages = pages.order_by('-live')

    if 'content_type' in request.GET:
        pagination_query_params['content_type'] = request.GET['content_type']

        app_label, model_name = request.GET['content_type'].split('.')

        try:
            selected_content_type = ContentType.objects.get_by_natural_key(app_label, model_name)
        except ContentType.DoesNotExist:
            raise Http404

        pages = pages.filter(content_type=selected_content_type)
    else:
        selected_content_type = None

    if 'q' in request.GET:
        form = SearchForm(request.GET)
        if form.is_valid():
            q = form.cleaned_data['q']
            pagination_query_params['q'] = q

            all_pages = all_pages.search(q, order_by_relevance=not ordering, operator='and')
            pages = pages.search(q, order_by_relevance=not ordering, operator='and')

            if pages.supports_facet:
                content_types = [
                    (ContentType.objects.get(id=content_type_id), count)
                    for content_type_id, count in all_pages.facet('content_type_id').items()
                ]

    else:
        form = SearchForm()

    paginator = Paginator(pages, per_page=20)
    pages = paginator.get_page(request.GET.get('p'))

    if request.is_ajax():
        return render(request, "wagtailadmin/pages/search_results.html", {
            'pages': pages,
            'all_pages': all_pages,
            'query_string': q,
            'content_types': content_types,
            'selected_content_type': selected_content_type,
            'ordering': ordering,
            'pagination_query_params': pagination_query_params.urlencode(),
        })
    else:
        return render(request, "wagtailadmin/pages/search.html", {
            'search_form': form,
            'pages': pages,
            'all_pages': all_pages,
            'query_string': q,
            'content_types': content_types,
            'selected_content_type': selected_content_type,
            'ordering': ordering,
            'pagination_query_params': pagination_query_params.urlencode(),
        })


def approve_moderation(request, revision_id):
    revision = get_object_or_404(PageRevision, id=revision_id)
    if not revision.page.permissions_for_user(request.user).can_publish():
        raise PermissionDenied

    if not revision.submitted_for_moderation:
        messages.error(request, _("The page '{0}' is not currently awaiting moderation.").format(revision.page.get_admin_display_title()))
        return redirect('wagtailadmin_home')

    if request.method == 'POST':
        revision.approve_moderation()

        message = _("Page '{0}' published.").format(revision.page.get_admin_display_title())
        buttons = []
        if revision.page.url is not None:
            buttons.append(messages.button(revision.page.url, _('View live'), new_window=True))
        buttons.append(messages.button(reverse('wagtailadmin_pages:edit', args=(revision.page.id,)), _('Edit')))
        messages.success(request, message, buttons=buttons)

        if not send_notification(revision.id, 'approved', request.user.pk):
            messages.error(request, _("Failed to send approval notifications"))

    return redirect('wagtailadmin_home')


def reject_moderation(request, revision_id):
    revision = get_object_or_404(PageRevision, id=revision_id)
    if not revision.page.permissions_for_user(request.user).can_publish():
        raise PermissionDenied

    if not revision.submitted_for_moderation:
        messages.error(request, _("The page '{0}' is not currently awaiting moderation.").format(revision.page.get_admin_display_title()))
        return redirect('wagtailadmin_home')

    if request.method == 'POST':
        revision.reject_moderation()
        messages.success(request, _("Page '{0}' rejected for publication.").format(revision.page.get_admin_display_title()), buttons=[
            messages.button(reverse('wagtailadmin_pages:edit', args=(revision.page.id,)), _('Edit'))
        ])
        if not send_notification(revision.id, 'rejected', request.user.pk):
            messages.error(request, _("Failed to send rejection notifications"))

    return redirect('wagtailadmin_home')


@require_GET
def preview_for_moderation(request, revision_id):
    revision = get_object_or_404(PageRevision, id=revision_id)
    if not revision.page.permissions_for_user(request.user).can_publish():
        raise PermissionDenied

    if not revision.submitted_for_moderation:
        messages.error(request, _("The page '{0}' is not currently awaiting moderation.").format(revision.page.get_admin_display_title()))
        return redirect('wagtailadmin_home')

    page = revision.as_page_object()

    return page.make_preview_request(request, page.default_preview_mode, extra_request_attrs={
        'revision_id': revision_id
    })


@require_POST
def lock(request, page_id):
    # Get the page
    page = get_object_or_404(Page, id=page_id).specific

    # Check permissions
    if not page.permissions_for_user(request.user).can_lock():
        raise PermissionDenied

    # Lock the page
    if not page.locked:
        page.locked = True
        page.save()

        messages.success(request, _("Page '{0}' is now locked.").format(page.get_admin_display_title()))

    # Redirect
    redirect_to = request.POST.get('next', None)
    if redirect_to and is_safe_url(url=redirect_to, allowed_hosts={request.get_host()}):
        return redirect(redirect_to)
    else:
        return redirect('wagtailadmin_explore', page.get_parent().id)


@require_POST
def unlock(request, page_id):
    # Get the page
    page = get_object_or_404(Page, id=page_id).specific

    # Check permissions
    if not page.permissions_for_user(request.user).can_lock():
        raise PermissionDenied

    # Unlock the page
    if page.locked:
        page.locked = False
        page.save()

        messages.success(request, _("Page '{0}' is now unlocked.").format(page.get_admin_display_title()))

    # Redirect
    redirect_to = request.POST.get('next', None)
    if redirect_to and is_safe_url(url=redirect_to, allowed_hosts={request.get_host()}):
        return redirect(redirect_to)
    else:
        return redirect('wagtailadmin_explore', page.get_parent().id)


@user_passes_test(user_has_any_page_permission)
def revisions_index(request, page_id):
    page = get_object_or_404(Page, id=page_id).specific

    # Get page ordering
    ordering = request.GET.get('ordering', '-created_at')
    if ordering not in ['created_at', '-created_at', ]:
        ordering = '-created_at'

    revisions = page.revisions.order_by(ordering)

    paginator = Paginator(revisions, per_page=20)
    revisions = paginator.get_page(request.GET.get('p'))

    return render(request, 'wagtailadmin/pages/revisions/index.html', {
        'page': page,
        'ordering': ordering,
        'pagination_query_params': "ordering=%s" % ordering,
        'revisions': revisions,
    })


def revisions_revert(request, page_id, revision_id):
    page = get_object_or_404(Page, id=page_id).specific
    page_perms = page.permissions_for_user(request.user)
    if not page_perms.can_edit():
        raise PermissionDenied

    revision = get_object_or_404(page.revisions, id=revision_id)
    revision_page = revision.as_page_object()

    content_type = ContentType.objects.get_for_model(page)
    page_class = content_type.model_class()

    edit_handler = page_class.get_edit_handler()
    edit_handler = edit_handler.bind_to(instance=revision_page,
                                        request=request)
    form_class = edit_handler.get_form_class()

    form = form_class(instance=revision_page)
    edit_handler = edit_handler.bind_to(form=form)

    user_avatar = render_to_string('wagtailadmin/shared/user_avatar.html', {'user': revision.user})

    messages.warning(request, mark_safe(
        _("You are viewing a previous revision of this page from <b>%(created_at)s</b> by %(user)s") % {
            'created_at': revision.created_at.strftime("%d %b %Y %H:%M"),
            'user': user_avatar,
        }
    ))

    return render(request, 'wagtailadmin/pages/edit.html', {
        'page': page,
        'revision': revision,
        'is_revision': True,
        'content_type': content_type,
        'edit_handler': edit_handler,
        'errors_debug': None,
        'action_menu': PageActionMenu(request, view='revisions_revert', page=page),
        'preview_modes': page.preview_modes,
        'form': form,  # Used in unit tests
    })


@user_passes_test(user_has_any_page_permission)
def revisions_view(request, page_id, revision_id):
    page = get_object_or_404(Page, id=page_id).specific

    perms = page.permissions_for_user(request.user)
    if not (perms.can_publish() or perms.can_edit()):
        raise PermissionDenied

    revision = get_object_or_404(page.revisions, id=revision_id)
    revision_page = revision.as_page_object()

    return revision_page.make_preview_request(request, page.default_preview_mode)


def revisions_compare(request, page_id, revision_id_a, revision_id_b):
    page = get_object_or_404(Page, id=page_id).specific

    # Get revision to compare from
    if revision_id_a == 'live':
        if not page.live:
            raise Http404

        revision_a = page
        revision_a_heading = _("Live")
    elif revision_id_a == 'earliest':
        revision_a = page.revisions.order_by('created_at', 'id').first()
        if revision_a:
            revision_a = revision_a.as_page_object()
            revision_a_heading = _("Earliest")
        else:
            raise Http404
    else:
        revision_a = get_object_or_404(page.revisions, id=revision_id_a).as_page_object()
        revision_a_heading = str(get_object_or_404(page.revisions, id=revision_id_a).created_at)

    # Get revision to compare to
    if revision_id_b == 'live':
        if not page.live:
            raise Http404

        revision_b = page
        revision_b_heading = _("Live")
    elif revision_id_b == 'latest':
        revision_b = page.revisions.order_by('created_at', 'id').last()
        if revision_b:
            revision_b = revision_b.as_page_object()
            revision_b_heading = _("Latest")
        else:
            raise Http404
    else:
        revision_b = get_object_or_404(page.revisions, id=revision_id_b).as_page_object()
        revision_b_heading = str(get_object_or_404(page.revisions, id=revision_id_b).created_at)

    comparison = page.get_edit_handler().get_comparison()
    comparison = [comp(revision_a, revision_b) for comp in comparison]
    comparison = [comp for comp in comparison if comp.has_changed()]

    return render(request, 'wagtailadmin/pages/revisions/compare.html', {
        'page': page,
        'revision_a_heading': revision_a_heading,
        'revision_a': revision_a,
        'revision_b_heading': revision_b_heading,
        'revision_b': revision_b,
        'comparison': comparison,
    })


def revisions_unschedule(request, page_id, revision_id):
    page = get_object_or_404(Page, id=page_id).specific

    user_perms = UserPagePermissionsProxy(request.user)
    if not user_perms.for_page(page).can_unschedule():
        raise PermissionDenied

    revision = get_object_or_404(page.revisions, id=revision_id)

    next_url = get_valid_next_url_from_request(request)

    subtitle = _('revision {0} of "{1}"').format(revision.id, page.get_admin_display_title())

    if request.method == 'POST':
        revision.approved_go_live_at = None
        revision.save(update_fields=['approved_go_live_at'])

        messages.success(request, _('Revision {0} of "{1}" unscheduled.').format(revision.id, page.get_admin_display_title()), buttons=[
            messages.button(reverse('wagtailadmin_pages:edit', args=(page.id,)), _('Edit'))
        ])

        if next_url:
            return redirect(next_url)
        return redirect('wagtailadmin_pages:revisions_index', page.id)

    return render(request, 'wagtailadmin/pages/revisions/confirm_unschedule.html', {
        'page': page,
        'revision': revision,
        'next': next_url,
        'subtitle': subtitle
    })



import locale

from pytz import country_timezones
import six

from .maps import tz_cities


def get_country_timezones(country_code):
    """
    Retrieves the timezones for a given country, sorted in alphabetical order
    """

    tz_list = []

    if country_code in country_timezones:
        tzs = country_timezones[country_code]
        tz_list = [(t, tz_cities[t]) for t in tzs]
        tz_list.sort(lambda x, y: locale.strcoll(x[1], y[1]))

    return tz_list


def get_country_code_from_tz(tz):
    """
    Retrieves the country matching a given timezone
    """

    for c, t in six.iteritems(country_timezones):
        if tz in t:
            return c
    return None



__version_info__ = (2016, 6, 4, 'final', 0)


def get_version(version=__version_info__):

    dev_st = {'alpha': 'a', 'beta': 'b', 'rc': 'c', 'final': ''}

    assert len(version) == 5
    assert version[3] in dev_st.keys()

    n = 2 + (version[2] != 0)
    version_str = '.'.join([str(v) for v in version[:n]])

    if version[3] == 'final':
        return version_str

    if version[3:] == ('alpha', 0):
        return '%s.dev%s' % (version_str, get_hg_chgset())
    else:
        return ''.join((version_str, dev_st[version[3]], str(version[4])))


def get_hg_chgset():
    import subprocess

    try:
        # python 3
        DEVNULL = subprocess.DEVNULL
    except AttributeError:
        import os
        DEVNULL = open(os.devnull, 'wb')

    try:
        return subprocess.check_output(['hg', 'id', '-i'],
                                       stderr=DEVNULL).strip()
    except:
        return '?'


__version__ = get_version()



import hashlib
import os.path
from contextlib import contextmanager

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.dispatch import Signal
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from taggit.managers import TaggableManager

from wagtail.admin.models import get_object_usage
from wagtail.core.models import CollectionMember
from wagtail.search import index
from wagtail.search.queryset import SearchableQuerySetMixin


class DocumentQuerySet(SearchableQuerySetMixin, models.QuerySet):
    pass


class AbstractDocument(CollectionMember, index.Indexed, models.Model):
    title = models.CharField(max_length=255, verbose_name=_('title'))
    file = models.FileField(upload_to='documents', verbose_name=_('file'))
    created_at = models.DateTimeField(verbose_name=_('created at'), auto_now_add=True)
    uploaded_by_user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_('uploaded by user'),
        null=True,
        blank=True,
        editable=False,
        on_delete=models.SET_NULL
    )

    tags = TaggableManager(help_text=None, blank=True, verbose_name=_('tags'))

    file_size = models.PositiveIntegerField(null=True, editable=False)
    # A SHA-1 hash of the file contents
    file_hash = models.CharField(max_length=40, blank=True, editable=False)

    objects = DocumentQuerySet.as_manager()

    search_fields = CollectionMember.search_fields + [
        index.SearchField('title', partial_match=True, boost=10),
        index.AutocompleteField('title'),
        index.FilterField('title'),
        index.RelatedFields('tags', [
            index.SearchField('name', partial_match=True, boost=10),
            index.AutocompleteField('name'),
        ]),
        index.FilterField('uploaded_by_user'),
    ]

    def is_stored_locally(self):
        """
        Returns True if the image is hosted on the local filesystem
        """
        try:
            self.file.path

            return True
        except NotImplementedError:
            return False

    @contextmanager
    def open_file(self):
        # Open file if it is closed
        close_file = False
        f = self.file

        if f.closed:
            # Reopen the file
            if self.is_stored_locally():
                f.open('rb')
            else:
                # Some external storage backends don't allow reopening
                # the file. Get a fresh file instance. #1397
                storage = self._meta.get_field('file').storage
                f = storage.open(f.name, 'rb')

            close_file = True

        # Seek to beginning
        f.seek(0)

        try:
            yield f
        finally:
            if close_file:
                f.close()

    def get_file_size(self):
        if self.file_size is None:
            try:
                self.file_size = self.file.size
            except Exception:
                # File doesn't exist
                return

            self.save(update_fields=['file_size'])

        return self.file_size

    def _set_file_hash(self, file_contents):
        self.file_hash = hashlib.sha1(file_contents).hexdigest()

    def get_file_hash(self):
        if self.file_hash == '':
            with self.open_file() as f:
                self._set_file_hash(f.read())

            self.save(update_fields=['file_hash'])

        return self.file_hash

    def __str__(self):
        return self.title

    @property
    def filename(self):
        return os.path.basename(self.file.name)

    @property
    def file_extension(self):
        return os.path.splitext(self.filename)[1][1:]

    @property
    def url(self):
        if getattr(settings, 'WAGTAILDOCS_SERVE_METHOD', None) == 'direct':
            try:
                return self.file.url
            except NotImplementedError:
                # backend does not provide a url, so fall back on the serve view
                pass

        return reverse('wagtaildocs_serve', args=[self.id, self.filename])

    def get_usage(self):
        return get_object_usage(self)

    @property
    def usage_url(self):
        return reverse('wagtaildocs:document_usage',
                       args=(self.id,))

    def is_editable_by_user(self, user):
        from wagtail.documents.permissions import permission_policy
        return permission_policy.user_has_permission_for_instance(user, 'change', self)

    class Meta:
        abstract = True
        verbose_name = _('document')
        verbose_name_plural = _('documents')


class Document(AbstractDocument):
    admin_form_fields = (
        'title',
        'file',
        'collection',
        'tags'
    )


def get_document_model():
    """
    Get the document model from the ``WAGTAILDOCS_DOCUMENT_MODEL`` setting.
    Defauts to the standard :class:`~wagtail.documents.models.Document` model
    if no custom model is defined.
    """
    from django.conf import settings
    from django.apps import apps

    try:
        app_label, model_name = settings.WAGTAILDOCS_DOCUMENT_MODEL.split('.')
    except AttributeError:
        return Document
    except ValueError:
        raise ImproperlyConfigured("WAGTAILDOCS_DOCUMENT_MODEL must be of the form 'app_label.model_name'")

    document_model = apps.get_model(app_label, model_name)
    if document_model is None:
        raise ImproperlyConfigured(
            "WAGTAILDOCS_DOCUMENT_MODEL refers to model '%s' that has not been installed" %
            settings.WAGTAILDOCS_DOCUMENT_MODEL
        )
    return document_model


document_served = Signal(providing_args=['request'])



import sys

from wagtail.utils.deprecation import MovedDefinitionHandler, RemovedInWagtail29Warning


MOVED_DEFINITIONS = {
    'WAGTAILADMIN_PROVIDED_LANGUAGES': 'wagtail.admin.localization',
    'get_js_translation_strings': 'wagtail.admin.localization',
    'get_available_admin_languages': 'wagtail.admin.localization',
    'get_available_admin_time_zones': 'wagtail.admin.localization',

    'get_object_usage': 'wagtail.admin.models',
    'popular_tags_for_model': 'wagtail.admin.models',

    'users_with_page_permission': 'wagtail.admin.auth',
    'permission_denied': 'wagtail.admin.auth',
    'user_passes_test': 'wagtail.admin.auth',
    'permission_required': 'wagtail.admin.auth',
    'any_permission_required': 'wagtail.admin.auth',
    'PermissionPolicyChecker': 'wagtail.admin.auth',
    'user_has_any_page_permission': 'wagtail.admin.auth',

    'send_mail': 'wagtail.admin.mail',
    'send_notification': 'wagtail.admin.mail',

    'get_site_for_user': 'wagtail.admin.navigation',
}

sys.modules[__name__] = MovedDefinitionHandler(sys.modules[__name__], MOVED_DEFINITIONS, RemovedInWagtail29Warning)



from django.test import TestCase
from django.test.utils import override_settings

from wagtail.admin import widgets
from wagtail.core.models import Page
from wagtail.tests.testapp.forms import AdminStarDateInput
from wagtail.tests.testapp.models import EventPage, SimplePage


class TestAdminPageChooserWidget(TestCase):
    def setUp(self):
        self.root_page = Page.objects.get(id=2)

        # Add child page
        self.child_page = SimplePage(
            title="foobarbaz",
            content="hello",
        )
        self.root_page.add_child(instance=self.child_page)

    def test_not_hidden(self):
        widget = widgets.AdminPageChooser()
        self.assertFalse(widget.is_hidden)

    def test_render_html(self):
        widget = widgets.AdminPageChooser()

        html = widget.render_html('test', None, {})
        self.assertInHTML("""<input name="test" type="hidden" />""", html)
        self.assertIn(">Choose a page<", html)

    def test_render_js_init(self):
        widget = widgets.AdminPageChooser()

        js_init = widget.render_js_init('test-id', 'test', None)
        self.assertEqual(js_init, "createPageChooser(\"test-id\", [\"wagtailcore.page\"], null, false, null);")

    def test_render_js_init_with_user_perm(self):
        widget = widgets.AdminPageChooser(user_perms='copy_to')

        js_init = widget.render_js_init('test-id', 'test', None)
        self.assertEqual(js_init, "createPageChooser(\"test-id\", [\"wagtailcore.page\"], null, false, \"copy_to\");")

    def test_render_html_with_value(self):
        widget = widgets.AdminPageChooser()

        html = widget.render_html('test', self.child_page, {})
        self.assertInHTML("""<input name="test" type="hidden" value="%d" />""" % self.child_page.id, html)

    def test_render_js_init_with_value(self):
        widget = widgets.AdminPageChooser()

        js_init = widget.render_js_init('test-id', 'test', self.child_page)
        self.assertEqual(
            js_init, "createPageChooser(\"test-id\", [\"wagtailcore.page\"], %d, false, null);" % self.root_page.id
        )

    # def test_render_html_init_with_content_type omitted as HTML does not
    # change when selecting a content type

    def test_render_js_init_with_target_model(self):
        widget = widgets.AdminPageChooser(target_models=[SimplePage])

        js_init = widget.render_js_init('test-id', 'test', None)
        self.assertEqual(js_init, "createPageChooser(\"test-id\", [\"tests.simplepage\"], null, false, null);")

        html = widget.render_html('test', self.child_page, {})
        self.assertIn(">Choose a page (Simple Page)<", html)

    def test_render_js_init_with_multiple_target_models(self):
        target_models = [SimplePage, EventPage]
        widget = widgets.AdminPageChooser(target_models=target_models)

        js_init = widget.render_js_init('test-id', 'test', None)
        self.assertEqual(
            js_init, "createPageChooser(\"test-id\", [\"tests.simplepage\", \"tests.eventpage\"], null, false, null);"
        )

        html = widget.render_html('test', self.child_page, {})
        self.assertIn(">Choose a page<", html)

    def test_render_js_init_with_can_choose_root(self):
        widget = widgets.AdminPageChooser(can_choose_root=True)

        js_init = widget.render_js_init('test-id', 'test', self.child_page)
        self.assertEqual(
            js_init, "createPageChooser(\"test-id\", [\"wagtailcore.page\"], %d, true, null);" % self.root_page.id
        )


class TestAdminDateInput(TestCase):

    def test_render_js_init(self):
        widget = widgets.AdminDateInput()

        html = widget.render('test', None, attrs={'id': 'test-id'})

        self.assertInHTML('<input type="text" name="test" autocomplete="new-date" id="test-id" />', html)

        # we should see the JS initialiser code:
        # initDateChooser("test-id", {"dayOfWeekStart": 0, "format": "Y-m-d"});
        # except that we can't predict the order of the config options
        self.assertIn('initDateChooser("test\\u002Did", {', html)
        self.assertIn('"dayOfWeekStart": 0', html)
        self.assertIn('"format": "Y-m-d"', html)

    def test_render_js_init_with_format(self):
        widget = widgets.AdminDateInput(format='%d.%m.%Y.')

        html = widget.render('test', None, attrs={'id': 'test-id'})
        self.assertIn(
            '"format": "d.m.Y."',
            html,
        )

    @override_settings(WAGTAIL_DATE_FORMAT='%d.%m.%Y.')
    def test_render_js_init_with_format_from_settings(self):
        widget = widgets.AdminDateInput()

        html = widget.render('test', None, attrs={'id': 'test-id'})
        self.assertIn(
            '"format": "d.m.Y."',
            html,
        )

    def test_media_inheritance(self):
        """
        Widgets inheriting from AdminDateInput should have their media definitions merged
        with AdminDateInput's
        """
        widget = AdminStarDateInput()
        media_html = str(widget.media)
        self.assertIn('wagtailadmin/js/date-time-chooser.js', media_html)
        self.assertIn('vendor/star_date.js', media_html)


class TestAdminDateTimeInput(TestCase):

    def test_render_js_init(self):
        widget = widgets.AdminDateTimeInput()

        html = widget.render('test', None, attrs={'id': 'test-id'})

        self.assertInHTML('<input type="text" name="test" autocomplete="new-date-time" id="test-id" />', html)

        # we should see the JS initialiser code:
        # initDateTimeChooser("test-id", {"dayOfWeekStart": 0, "format": "Y-m-d H:i"});
        # except that we can't predict the order of the config options
        self.assertIn('initDateTimeChooser("test\\u002Did", {', html)
        self.assertIn('"dayOfWeekStart": 0', html)
        self.assertIn('"format": "Y-m-d H:i"', html)

    def test_render_js_init_with_format(self):
        widget = widgets.AdminDateTimeInput(format='%d.%m.%Y. %H:%M')

        html = widget.render('test', None, attrs={'id': 'test-id'})
        self.assertIn(
            '"format": "d.m.Y. H:i"',
            html,
        )

    @override_settings(WAGTAIL_DATETIME_FORMAT='%d.%m.%Y. %H:%M')
    def test_render_js_init_with_format_from_settings(self):
        widget = widgets.AdminDateTimeInput()

        html = widget.render('test', None, attrs={'id': 'test-id'})
        self.assertIn(
            '"format": "d.m.Y. H:i"',
            html,
        )


class TestAdminTagWidget(TestCase):

    def get_js_init_params(self, html):
        """Returns a list of the params passed in to initTagField from the supplied HTML"""
        # Eg. ["'test\\u002Did'", "'/admin/tag\\u002Dautocomplete/'", 'true', 'null']
        start = 'initTagField('
        end = ');'
        items_after_init = html.split(start)[1]
        if items_after_init:
            params_raw = items_after_init.split(end)[0]
            if params_raw:
                return [part.strip() for part in params_raw.split(',')]
        return []


    def test_render_js_init_basic(self):
        """Chekcs that the 'initTagField' is correctly added to the inline script for tag widgets"""
        widget = widgets.AdminTagWidget()

        html = widget.render('tags', None, attrs={'id': 'alpha'})
        params = self.get_js_init_params(html)

        self.assertEqual(len(params), 4)
        self.assertEqual(params[0], "'alpha'")  # id
        self.assertEqual(params[1], "'/admin/tag\\u002Dautocomplete/'")  # autocomplete url
        self.assertEqual(params[2], 'true')  # tag_spaces_allowed
        self.assertEqual(params[3], 'null')  # tag_limit


    @override_settings(TAG_SPACES_ALLOWED=False)
    def test_render_js_init_no_spaces_allowed(self):
        """Chekcs that the 'initTagField' includes the correct value based on TAG_SPACES_ALLOWED in settings"""
        widget = widgets.AdminTagWidget()

        html = widget.render('tags', None, attrs={'id': 'alpha'})
        params = self.get_js_init_params(html)

        self.assertEqual(len(params), 4)
        self.assertEqual(params[2], 'false')  # tag_spaces_allowed
        self.assertEqual(params[3], 'null')  # tag_limit


    @override_settings(TAG_LIMIT=5)
    def test_render_js_init_with_tag_limit(self):
        """Chekcs that the 'initTagField' includes the correct value based on TAG_LIMIT in settings"""
        widget = widgets.AdminTagWidget()

        html = widget.render('tags', None, attrs={'id': 'alpha'})
        params = self.get_js_init_params(html)

        self.assertEqual(len(params), 4)
        self.assertEqual(params[2], 'true')  # tag_spaces_allowed
        self.assertEqual(params[3], '5')  # tag_limit



import json
import os

from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.shortcuts import render
from django.utils.text import slugify
from django.utils.translation import ugettext_lazy as _
from unidecode import unidecode

from wagtail.admin.edit_handlers import FieldPanel
from wagtail.admin.mail import send_mail
from wagtail.core.models import Orderable, Page

from .forms import FormBuilder, WagtailAdminFormPageForm
from .views import SubmissionsListView

FORM_FIELD_CHOICES = (
    ('singleline', _('Single line text')),
    ('multiline', _('Multi-line text')),
    ('email', _('Email')),
    ('number', _('Number')),
    ('url', _('URL')),
    ('checkbox', _('Checkbox')),
    ('checkboxes', _('Checkboxes')),
    ('dropdown', _('Drop down')),
    ('multiselect', _('Multiple select')),
    ('radio', _('Radio buttons')),
    ('date', _('Date')),
    ('datetime', _('Date/time')),
    ('hidden', _('Hidden field')),
)


class AbstractFormSubmission(models.Model):
    """
    Data for a form submission.

    You can create custom submission model based on this abstract model.
    For example, if you need to save additional data or a reference to a user.
    """

    form_data = models.TextField()
    page = models.ForeignKey(Page, on_delete=models.CASCADE)

    submit_time = models.DateTimeField(verbose_name=_('submit time'), auto_now_add=True)

    def get_data(self):
        """
        Returns dict with form data.

        You can override this method to add additional data.
        """
        form_data = json.loads(self.form_data)
        form_data.update({
            'submit_time': self.submit_time,
        })

        return form_data

    def __str__(self):
        return self.form_data

    class Meta:
        abstract = True
        verbose_name = _('form submission')


class FormSubmission(AbstractFormSubmission):
    """Data for a Form submission."""


class AbstractFormField(Orderable):
    """
    Database Fields required for building a Django Form field.
    """

    label = models.CharField(
        verbose_name=_('label'),
        max_length=255,
        help_text=_('The label of the form field')
    )
    field_type = models.CharField(verbose_name=_('field type'), max_length=16, choices=FORM_FIELD_CHOICES)
    required = models.BooleanField(verbose_name=_('required'), default=True)
    choices = models.TextField(
        verbose_name=_('choices'),
        blank=True,
        help_text=_('Comma separated list of choices. Only applicable in checkboxes, radio and dropdown.')
    )
    default_value = models.CharField(
        verbose_name=_('default value'),
        max_length=255,
        blank=True,
        help_text=_('Default value. Comma separated values supported for checkboxes.')
    )
    help_text = models.CharField(verbose_name=_('help text'), max_length=255, blank=True)

    @property
    def clean_name(self):
        # unidecode will return an ascii string while slugify wants a
        # unicode string on the other hand, slugify returns a safe-string
        # which will be converted to a normal str
        return str(slugify(str(unidecode(self.label))))

    panels = [
        FieldPanel('label'),
        FieldPanel('help_text'),
        FieldPanel('required'),
        FieldPanel('field_type', classname="formbuilder-type"),
        FieldPanel('choices', classname="formbuilder-choices"),
        FieldPanel('default_value', classname="formbuilder-default"),
    ]

    class Meta:
        abstract = True
        ordering = ['sort_order']


class AbstractForm(Page):
    """
    A Form Page. Pages implementing a form should inherit from it
    """

    base_form_class = WagtailAdminFormPageForm

    form_builder = FormBuilder

    submissions_list_view_class = SubmissionsListView

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not hasattr(self, 'landing_page_template'):
            name, ext = os.path.splitext(self.template)
            self.landing_page_template = name + '_landing' + ext

    class Meta:
        abstract = True

    def get_form_fields(self):
        """
        Form page expects `form_fields` to be declared.
        If you want to change backwards relation name,
        you need to override this method.
        """

        return self.form_fields.all()

    def get_data_fields(self):
        """
        Returns a list of tuples with (field_name, field_label).
        """

        data_fields = [
            ('submit_time', _('Submission date')),
        ]
        data_fields += [
            (field.clean_name, field.label)
            for field in self.get_form_fields()
        ]

        return data_fields

    def get_form_class(self):
        fb = self.form_builder(self.get_form_fields())
        return fb.get_form_class()

    def get_form_parameters(self):
        return {}

    def get_form(self, *args, **kwargs):
        form_class = self.get_form_class()
        form_params = self.get_form_parameters()
        form_params.update(kwargs)

        return form_class(*args, **form_params)

    def get_landing_page_template(self, request, *args, **kwargs):
        return self.landing_page_template

    def get_submission_class(self):
        """
        Returns submission class.

        You can override this method to provide custom submission class.
        Your class must be inherited from AbstractFormSubmission.
        """

        return FormSubmission

    def process_form_submission(self, form):
        """
        Accepts form instance with submitted data, user and page.
        Creates submission instance.

        You can override this method if you want to have custom creation logic.
        For example, if you want to save reference to a user.
        """

        return self.get_submission_class().objects.create(
            form_data=json.dumps(form.cleaned_data, cls=DjangoJSONEncoder),
            page=self,
        )

    def render_landing_page(self, request, form_submission=None, *args, **kwargs):
        """
        Renders the landing page.

        You can override this method to return a different HttpResponse as
        landing page. E.g. you could return a redirect to a separate page.
        """
        context = self.get_context(request)
        context['form_submission'] = form_submission
        return render(
            request,
            self.get_landing_page_template(request),
            context
        )

    def serve_submissions_list_view(self, request, *args, **kwargs):
        """
        Returns list submissions view for admin.

        `list_submissions_view_class` can bse set to provide custom view class.
        Your class must be inherited from SubmissionsListView.
        """
        view = self.submissions_list_view_class.as_view()
        return view(request, form_page=self, *args, **kwargs)

    def serve(self, request, *args, **kwargs):
        if request.method == 'POST':
            form = self.get_form(request.POST, request.FILES, page=self, user=request.user)

            if form.is_valid():
                form_submission = self.process_form_submission(form)
                return self.render_landing_page(request, form_submission, *args, **kwargs)
        else:
            form = self.get_form(page=self, user=request.user)

        context = self.get_context(request)
        context['form'] = form
        return render(
            request,
            self.get_template(request),
            context
        )

    preview_modes = [
        ('form', _('Form')),
        ('landing', _('Landing page')),
    ]

    def serve_preview(self, request, mode):
        if mode == 'landing':
            request.is_preview = True
            return self.render_landing_page(request)
        else:
            return super().serve_preview(request, mode)


class AbstractEmailForm(AbstractForm):
    """
    A Form Page that sends email. Pages implementing a form to be send to an email should inherit from it
    """

    to_address = models.CharField(
        verbose_name=_('to address'), max_length=255, blank=True,
        help_text=_("Optional - form submissions will be emailed to these addresses. Separate multiple addresses by comma.")
    )
    from_address = models.CharField(verbose_name=_('from address'), max_length=255, blank=True)
    subject = models.CharField(verbose_name=_('subject'), max_length=255, blank=True)

    def process_form_submission(self, form):
        submission = super().process_form_submission(form)
        if self.to_address:
            self.send_mail(form)
        return submission

    def send_mail(self, form):
        addresses = [x.strip() for x in self.to_address.split(',')]
        content = []
        for field in form:
            value = field.value()
            if isinstance(value, list):
                value = ', '.join(value)
            content.append('{}: {}'.format(field.label, value))
        content = '\n'.join(content)
        send_mail(self.subject, content, addresses, self.from_address,)

    class Meta:
        abstract = True



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class WagtailCoreAppConfig(AppConfig):
    name = 'wagtail.core'
    label = 'wagtailcore'
    verbose_name = _("Wagtail core")

    def ready(self):
        from wagtail.core.signal_handlers import register_signal_handlers
        register_signal_handlers()



from django.contrib.auth.models import Permission
from django.test import TestCase
from django.urls import reverse

from wagtail.core.models import Page
from wagtail.tests.testapp.models import SimplePage
from wagtail.tests.utils import WagtailTestUtils


class TestLocking(TestCase, WagtailTestUtils):
    def setUp(self):
        # Find root page
        self.root_page = Page.objects.get(id=2)

        # Login
        self.user = self.login()

        # Create a page and submit it for moderation
        self.child_page = SimplePage(
            title="Hello world!",
            slug='hello-world',
            content="hello",
            live=False,
        )
        self.root_page.add_child(instance=self.child_page)

    def test_lock_post(self):
        response = self.client.post(reverse('wagtailadmin_pages:lock', args=(self.child_page.id, )))

        # Check response
        self.assertRedirects(response, reverse('wagtailadmin_explore', args=(self.root_page.id, )))

        # Check that the page is locked
        self.assertTrue(Page.objects.get(id=self.child_page.id).locked)

    def test_lock_get(self):
        response = self.client.get(reverse('wagtailadmin_pages:lock', args=(self.child_page.id, )))

        # Check response
        self.assertEqual(response.status_code, 405)

        # Check that the page is still unlocked
        self.assertFalse(Page.objects.get(id=self.child_page.id).locked)

    def test_lock_post_already_locked(self):
        # Lock the page
        self.child_page.locked = True
        self.child_page.save()

        response = self.client.post(reverse('wagtailadmin_pages:lock', args=(self.child_page.id, )))

        # Check response
        self.assertRedirects(response, reverse('wagtailadmin_explore', args=(self.root_page.id, )))

        # Check that the page is still locked
        self.assertTrue(Page.objects.get(id=self.child_page.id).locked)

    def test_lock_post_with_good_redirect(self):
        response = self.client.post(reverse('wagtailadmin_pages:lock', args=(self.child_page.id, )), {
            'next': reverse('wagtailadmin_pages:edit', args=(self.child_page.id, ))
        })

        # Check response
        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=(self.child_page.id, )))

        # Check that the page is locked
        self.assertTrue(Page.objects.get(id=self.child_page.id).locked)

    def test_lock_post_with_bad_redirect(self):
        response = self.client.post(reverse('wagtailadmin_pages:lock', args=(self.child_page.id, )), {
            'next': 'http://www.google.co.uk'
        })

        # Check response
        self.assertRedirects(response, reverse('wagtailadmin_explore', args=(self.root_page.id, )))

        # Check that the page is locked
        self.assertTrue(Page.objects.get(id=self.child_page.id).locked)

    def test_lock_post_bad_page(self):
        response = self.client.post(reverse('wagtailadmin_pages:lock', args=(9999, )))

        # Check response
        self.assertEqual(response.status_code, 404)

        # Check that the page is still unlocked
        self.assertFalse(Page.objects.get(id=self.child_page.id).locked)

    def test_lock_post_bad_permissions(self):
        # Remove privileges from user
        self.user.is_superuser = False
        self.user.user_permissions.add(
            Permission.objects.get(content_type__app_label='wagtailadmin', codename='access_admin')
        )
        self.user.save()

        response = self.client.post(reverse('wagtailadmin_pages:lock', args=(self.child_page.id, )))

        # Check response
        self.assertEqual(response.status_code, 403)

        # Check that the page is still unlocked
        self.assertFalse(Page.objects.get(id=self.child_page.id).locked)

    def test_unlock_post(self):
        # Lock the page
        self.child_page.locked = True
        self.child_page.save()

        response = self.client.post(reverse('wagtailadmin_pages:unlock', args=(self.child_page.id, )))

        # Check response
        self.assertRedirects(response, reverse('wagtailadmin_explore', args=(self.root_page.id, )))

        # Check that the page is unlocked
        self.assertFalse(Page.objects.get(id=self.child_page.id).locked)

    def test_unlock_get(self):
        # Lock the page
        self.child_page.locked = True
        self.child_page.save()

        response = self.client.get(reverse('wagtailadmin_pages:unlock', args=(self.child_page.id, )))

        # Check response
        self.assertEqual(response.status_code, 405)

        # Check that the page is still locked
        self.assertTrue(Page.objects.get(id=self.child_page.id).locked)

    def test_unlock_post_already_unlocked(self):
        response = self.client.post(reverse('wagtailadmin_pages:unlock', args=(self.child_page.id, )))

        # Check response
        self.assertRedirects(response, reverse('wagtailadmin_explore', args=(self.root_page.id, )))

        # Check that the page is still unlocked
        self.assertFalse(Page.objects.get(id=self.child_page.id).locked)

    def test_unlock_post_with_good_redirect(self):
        # Lock the page
        self.child_page.locked = True
        self.child_page.save()

        response = self.client.post(reverse('wagtailadmin_pages:unlock', args=(self.child_page.id, )), {
            'next': reverse('wagtailadmin_pages:edit', args=(self.child_page.id, ))
        })

        # Check response
        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=(self.child_page.id, )))

        # Check that the page is unlocked
        self.assertFalse(Page.objects.get(id=self.child_page.id).locked)

    def test_unlock_post_with_bad_redirect(self):
        # Lock the page
        self.child_page.locked = True
        self.child_page.save()

        response = self.client.post(reverse('wagtailadmin_pages:unlock', args=(self.child_page.id, )), {
            'next': 'http://www.google.co.uk'
        })

        # Check response
        self.assertRedirects(response, reverse('wagtailadmin_explore', args=(self.root_page.id, )))

        # Check that the page is unlocked
        self.assertFalse(Page.objects.get(id=self.child_page.id).locked)

    def test_unlock_post_bad_page(self):
        # Lock the page
        self.child_page.locked = True
        self.child_page.save()

        response = self.client.post(reverse('wagtailadmin_pages:unlock', args=(9999, )))

        # Check response
        self.assertEqual(response.status_code, 404)

        # Check that the page is still locked
        self.assertTrue(Page.objects.get(id=self.child_page.id).locked)

    def test_unlock_post_bad_permissions(self):
        # Remove privileges from user
        self.user.is_superuser = False
        self.user.user_permissions.add(
            Permission.objects.get(content_type__app_label='wagtailadmin', codename='access_admin')
        )
        self.user.save()

        # Lock the page
        self.child_page.locked = True
        self.child_page.save()

        response = self.client.post(reverse('wagtailadmin_pages:unlock', args=(self.child_page.id, )))

        # Check response
        self.assertEqual(response.status_code, 403)

        # Check that the page is still locked
        self.assertTrue(Page.objects.get(id=self.child_page.id).locked)



import itertools
import json
from functools import total_ordering

from django import forms
from django.conf import settings
from django.forms import widgets
from django.forms.utils import flatatt
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.formats import get_format
from django.utils.functional import cached_property
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _
from taggit.forms import TagWidget

from wagtail.admin.datetimepicker import to_datetimepicker_format
from wagtail.admin.staticfiles import versioned_static
from wagtail.core import hooks
from wagtail.core.models import Page
from wagtail.utils.widgets import WidgetWithScript

DEFAULT_DATE_FORMAT = '%Y-%m-%d'
DEFAULT_DATETIME_FORMAT = '%Y-%m-%d %H:%M'


class AdminAutoHeightTextInput(widgets.Textarea):
    template_name = 'wagtailadmin/widgets/auto_height_text_input.html'

    def __init__(self, attrs=None):
        # Use more appropriate rows default, given autoheight will alter this anyway
        default_attrs = {'rows': '1'}
        if attrs:
            default_attrs.update(attrs)

        super().__init__(default_attrs)


class AdminDateInput(widgets.DateInput):
    template_name = 'wagtailadmin/widgets/date_input.html'

    def __init__(self, attrs=None, format=None):
        default_attrs = {'autocomplete': 'new-date'}
        fmt = format
        if attrs:
            default_attrs.update(attrs)
        if fmt is None:
            fmt = getattr(settings, 'WAGTAIL_DATE_FORMAT', DEFAULT_DATE_FORMAT)
        self.js_format = to_datetimepicker_format(fmt)
        super().__init__(attrs=default_attrs, format=fmt)

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)

        config = {
            'dayOfWeekStart': get_format('FIRST_DAY_OF_WEEK'),
            'format': self.js_format,
        }
        context['widget']['config_json'] = json.dumps(config)

        return context

    @property
    def media(self):
        return forms.Media(js=[
            versioned_static('wagtailadmin/js/date-time-chooser.js'),
        ])


class AdminTimeInput(widgets.TimeInput):
    template_name = 'wagtailadmin/widgets/time_input.html'

    def __init__(self, attrs=None, format='%H:%M'):
        default_attrs = {'autocomplete': 'new-time'}
        if attrs:
            default_attrs.update(attrs)
        super().__init__(attrs=default_attrs, format=format)

    @property
    def media(self):
        return forms.Media(js=[
            versioned_static('wagtailadmin/js/date-time-chooser.js'),
        ])


class AdminDateTimeInput(widgets.DateTimeInput):
    template_name = 'wagtailadmin/widgets/datetime_input.html'

    def __init__(self, attrs=None, format=None):
        default_attrs = {'autocomplete': 'new-date-time'}
        fmt = format
        if attrs:
            default_attrs.update(attrs)
        if fmt is None:
            fmt = getattr(settings, 'WAGTAIL_DATETIME_FORMAT', DEFAULT_DATETIME_FORMAT)
        self.js_format = to_datetimepicker_format(fmt)
        super().__init__(attrs=default_attrs, format=fmt)

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)

        config = {
            'dayOfWeekStart': get_format('FIRST_DAY_OF_WEEK'),
            'format': self.js_format,
        }
        context['widget']['config_json'] = json.dumps(config)

        return context

    @property
    def media(self):
        return forms.Media(js=[
            versioned_static('wagtailadmin/js/date-time-chooser.js'),
        ])


class AdminTagWidget(TagWidget):
    template_name = 'wagtailadmin/widgets/tag_widget.html'

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        context['widget']['autocomplete_url'] = reverse('wagtailadmin_tag_autocomplete')
        context['widget']['tag_spaces_allowed'] = getattr(settings, 'TAG_SPACES_ALLOWED', True)
        context['widget']['tag_limit'] = getattr(settings, 'TAG_LIMIT', None)

        return context


class AdminChooser(WidgetWithScript, widgets.Input):
    input_type = 'hidden'
    choose_one_text = _("Choose an item")
    choose_another_text = _("Choose another item")
    clear_choice_text = _("Clear choice")
    link_to_chosen_text = _("Edit this item")
    show_edit_link = True

    # when looping over form fields, this one should appear in visible_fields, not hidden_fields
    # despite the underlying input being type="hidden"
    is_hidden = False

    def get_instance(self, model_class, value):
        # helper method for cleanly turning 'value' into an instance object
        if value is None:
            return None

        try:
            return model_class.objects.get(pk=value)
        except model_class.DoesNotExist:
            return None

    def get_instance_and_id(self, model_class, value):
        if value is None:
            return (None, None)
        elif isinstance(value, model_class):
            return (value, value.pk)
        else:
            try:
                return (model_class.objects.get(pk=value), value)
            except model_class.DoesNotExist:
                return (None, None)

    def value_from_datadict(self, data, files, name):
        # treat the empty string as None
        result = super().value_from_datadict(data, files, name)
        if result == '':
            return None
        else:
            return result

    def __init__(self, **kwargs):
        # allow choose_one_text / choose_another_text to be overridden per-instance
        if 'choose_one_text' in kwargs:
            self.choose_one_text = kwargs.pop('choose_one_text')
        if 'choose_another_text' in kwargs:
            self.choose_another_text = kwargs.pop('choose_another_text')
        if 'clear_choice_text' in kwargs:
            self.clear_choice_text = kwargs.pop('clear_choice_text')
        if 'link_to_chosen_text' in kwargs:
            self.link_to_chosen_text = kwargs.pop('link_to_chosen_text')
        if 'show_edit_link' in kwargs:
            self.show_edit_link = kwargs.pop('show_edit_link')
        super().__init__(**kwargs)


class AdminPageChooser(AdminChooser):
    choose_one_text = _('Choose a page')
    choose_another_text = _('Choose another page')
    link_to_chosen_text = _('Edit this page')

    def __init__(self, target_models=None, can_choose_root=False, user_perms=None, **kwargs):
        super().__init__(**kwargs)

        if target_models:
            model_names = [model._meta.verbose_name.title() for model in target_models if model is not Page]
            if len(model_names) == 1:
                self.choose_one_text += ' (' + model_names[0] + ')'

        self.user_perms = user_perms
        self.target_models = list(target_models or [Page])
        self.can_choose_root = can_choose_root

    def _get_lowest_common_page_class(self):
        """
        Return a Page class that is an ancestor for all Page classes in
        ``target_models``, and is also a concrete Page class itself.
        """
        if len(self.target_models) == 1:
            # Shortcut for a single page type
            return self.target_models[0]
        else:
            return Page

    def render_html(self, name, value, attrs):
        model_class = self._get_lowest_common_page_class()

        instance, value = self.get_instance_and_id(model_class, value)

        original_field_html = super().render_html(name, value, attrs)

        return render_to_string("wagtailadmin/widgets/page_chooser.html", {
            'widget': self,
            'original_field_html': original_field_html,
            'attrs': attrs,
            'value': value,
            'page': instance,
        })

    def render_js_init(self, id_, name, value):
        if isinstance(value, Page):
            page = value
        else:
            # Value is an ID look up object
            model_class = self._get_lowest_common_page_class()
            page = self.get_instance(model_class, value)

        parent = page.get_parent() if page else None

        return "createPageChooser({id}, {model_names}, {parent}, {can_choose_root}, {user_perms});".format(
            id=json.dumps(id_),
            model_names=json.dumps([
                '{app}.{model}'.format(
                    app=model._meta.app_label,
                    model=model._meta.model_name)
                for model in self.target_models
            ]),
            parent=json.dumps(parent.id if parent else None),
            can_choose_root=('true' if self.can_choose_root else 'false'),
            user_perms=json.dumps(self.user_perms),
        )

    @property
    def media(self):
        return forms.Media(js=[
            versioned_static('wagtailadmin/js/page-chooser-modal.js'),
            versioned_static('wagtailadmin/js/page-chooser.js'),
        ])


@total_ordering
class Button:
    show = True

    def __init__(self, label, url, classes=set(), attrs={}, priority=1000):
        self.label = label
        self.url = url
        self.classes = classes
        self.attrs = attrs.copy()
        self.priority = priority

    def render(self):
        attrs = {'href': self.url, 'class': ' '.join(sorted(self.classes))}
        attrs.update(self.attrs)
        return format_html('<a{}>{}</a>', flatatt(attrs), self.label)

    def __str__(self):
        return self.render()

    def __repr__(self):
        return '<Button: {}>'.format(self.label)

    def __lt__(self, other):
        if not isinstance(other, Button):
            return NotImplemented
        return (self.priority, self.label) < (other.priority, other.label)

    def __eq__(self, other):
        if not isinstance(other, Button):
            return NotImplemented
        return (self.label == other.label
                and self.url == other.url
                and self.classes == other.classes
                and self.attrs == other.attrs
                and self.priority == other.priority)


class PageListingButton(Button):
    def __init__(self, label, url, classes=set(), **kwargs):
        classes = {'button', 'button-small', 'button-secondary'} | set(classes)
        super().__init__(label, url, classes=classes, **kwargs)


class BaseDropdownMenuButton(Button):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, url=None, **kwargs)

    @cached_property
    def dropdown_buttons(self):
        raise NotImplementedError

    def render(self):
        return render_to_string(self.template_name, {
            'buttons': self.dropdown_buttons,
            'label': self.label,
            'title': self.attrs.get('title'),
            'is_parent': self.is_parent})


class ButtonWithDropdownFromHook(BaseDropdownMenuButton):
    template_name = 'wagtailadmin/pages/listing/_button_with_dropdown.html'

    def __init__(self, label, hook_name, page, page_perms, is_parent, **kwargs):
        self.hook_name = hook_name
        self.page = page
        self.page_perms = page_perms
        self.is_parent = is_parent

        super().__init__(label, **kwargs)

    @property
    def show(self):
        return bool(self.dropdown_buttons)

    @cached_property
    def dropdown_buttons(self):
        button_hooks = hooks.get_hooks(self.hook_name)
        return sorted(itertools.chain.from_iterable(
            hook(self.page, self.page_perms, self.is_parent)
            for hook in button_hooks))



import sys
from wagtail.utils.deprecation import MovedDefinitionHandler, RemovedInWagtail29Warning

MOVED_DEFINITIONS = {
    'reject_request': 'wagtail.admin.auth',
    'require_admin_access': 'wagtail.admin.auth',
}

sys.modules[__name__] = MovedDefinitionHandler(sys.modules[__name__], MOVED_DEFINITIONS, RemovedInWagtail29Warning)
